# coding: utf-8
# Author: Felix Fontein <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2021

import asyncio
import os
import os.path
import re
import typing as t

import aiofiles
import asyncio_pool

from . import app_context
from .logging import log
from .yaml import load_yaml_file


mlog = log.fields(mod=__name__)

_RST_LABEL_DEFINITION = re.compile(r'''^\.\. _([^:]+):''')


class ExtraDocsIndexError(Exception):
    pass


class Section:
    title: str
    toctree: t.List[str]

    def __init__(self, title: str, toctree: t.List[str]):
        self.title = title
        self.toctree = toctree


#: A tuple consisting of a list of sections and a list of RST documents as tuples
#: (absolute path to source file, relative path in collection's docs directory).
CollectionExtraDocsInfoT = t.Tuple[t.List[Section], t.List[t.Tuple[str, str]]]


def find_extra_docs(path_to_collection: str) -> t.List[t.Tuple[str, str]]:
    '''Enumerate all extra docs RST files for a collection path.

    :arg path_to_collection: Path to a collection.
    :arg collection_name: Dotted collection name.
    :returns: A list of tuples (absolute path, relative path in docs/docsite/rst)
    '''
    docs_dir = os.path.join(path_to_collection, 'docs', 'docsite', 'rst')
    if not os.path.isdir(docs_dir):
        return []
    result = []
    for dirname, dirs, files in os.walk(docs_dir):
        for file in files:
            if file.endswith('.rst'):
                path = os.path.join(dirname, file)
                result.append((path, os.path.normpath(os.path.relpath(path, docs_dir))))
    return result


def get_label_prefix(collection_name: str) -> str:
    '''Create RST label prefix for the given collection name.

    :arg collection_name: Dotted collection name.
    :returns: A RST label prefix
    '''
    return f'ansible_collections.{collection_name}.docsite.'


def lint_required_conditions(content: str, collection_name: str
                             ) -> t.Tuple[t.List[str], t.List[t.Tuple[int, int, str]]]:
    '''Check a extra docs RST file's content for whether it satisfied the required conditions.

    :arg content: Content of a RST document.
    :arg collection_name: Dotted collection name.
    :returns: A tuple consisting of a list of RST labels and a list of error messages
              (with line and column numbers).
    '''
    labels: t.Set[str] = set()
    errors: t.List[t.Tuple[int, int, str]] = []
    label_prefix = get_label_prefix(collection_name)
    # Check label definitions
    for row, line in enumerate(content.splitlines()):
        m = _RST_LABEL_DEFINITION.match(line)
        if m:
            label = m.group(1)
            if not label.startswith(label_prefix):
                errors.append((
                    row + 1,
                    0,
                    f'Label "{label}" does not start with expected prefix "{label_prefix}"'))
            else:
                labels.add(label)
    return sorted(labels), errors


def load_toctree(yaml_section: t.Dict[str, t.Any], section_index: int = 0
                 ) -> t.Tuple[t.List[str], t.List[str]]:
    errors: t.List[str] = []
    toctree: t.List[str] = []
    if 'toctree' in yaml_section:
        if not isinstance(yaml_section['toctree'], list):
            errors.append(
                f'Toctree entry in section #{section_index} is not a list')
            return toctree, errors

        for toctree_index, toctree_name in enumerate(yaml_section['toctree']):
            if not isinstance(toctree_name, str):
                errors.append(
                    f'Toctree entry #{toctree_index} in section #{section_index}'
                    f' is not a string')
                continue
            toctree.append(toctree_name)
    return toctree, errors


def load_section(yaml_section: t.Dict[str, t.Any], section_index: int = 0
                 ) -> t.Tuple[t.Optional[Section], t.List[str]]:
    errors: t.List[str] = []
    missing = False
    for required_key in ('title', ):
        if required_key not in yaml_section:
            errors.append(
                f'Section #{section_index} has no "{required_key}" entry')
            missing = True
    if missing:
        return None, errors
    toctree, toctree_errors = load_toctree(yaml_section, section_index)
    errors.extend(toctree_errors)
    if not toctree:
        errors.append(
            f'Section #{section_index} has no content')
        return None, errors
    return Section(yaml_section['title'], toctree), errors


def load_extra_docs_index(index_path: str) -> t.Tuple[t.List[Section], t.List[str]]:
    '''Load a collection's extra-docs.yml file.

    :arg index_path: Path to extra-docs.yml (does not need to exist).
    :returns: A tuple consisting of a list of sections and a list of error messages.
    :raises: ExtraDocsIndexError if extra-docs.yml does not exist
    '''
    sections: t.List[Section] = []
    errors: t.List[str] = []

    if not os.path.isfile(index_path):
        raise ExtraDocsIndexError('extra-docs.yml does not exist')

    try:
        index = load_yaml_file(index_path)
        if index.get('sections'):
            for section_index, yaml_section in enumerate(index['sections']):
                if not isinstance(yaml_section, dict):
                    errors.append(f'Section #{section_index} must be a mapping')
                    continue
                section, section_errors = load_section(yaml_section, section_index)
                if section is not None:
                    sections.append(section)
                errors.extend(section_errors)
    except Exception as exc:
        errors.append(str(exc))

    return sections, errors


async def load_collection_extra_docs(collection_name: str,
                                     collection_path: str,
                                     path_prefix: str = 'docsite'
                                     ) -> CollectionExtraDocsInfoT:
    '''Given a collection name and collection metadata, load extra docs data.

    :arg collection_name: Dotted collection name.
    :arg collection_path: Path to the collection.
    :arg path_prefix: Prefix to add to relative paths, and toctree entries.
    :returns: A tuple consisting of a list of sections and a list of RST documents as tuples
              (relative path in docs/docsite/rst, content).
    '''
    flog = mlog.fields(func='load_collection_extra_docs')
    flog.debug('Enter')

    index_path = os.path.join(collection_path, 'docs', 'docsite', 'extra-docs.yml')
    try:
        sections, dummy = load_extra_docs_index(index_path)
    except ExtraDocsIndexError:
        sections = []

    for section in sections:
        for i, toctree in enumerate(section.toctree):
            section.toctree[i] = f"{path_prefix}/{toctree}"
    documents = []
    for abs_path, rel_path in find_extra_docs(collection_path):
        try:
            # Load content
            async with aiofiles.open(abs_path, 'r', encoding='utf-8') as f:
                content = await f.read()

            # Lint content
            dummy, errors = lint_required_conditions(content, collection_name)

            # When no errors were found, add to output
            if not errors:
                documents.append((abs_path, os.path.join(path_prefix, rel_path)))
        except Exception:
            pass

    flog.debug('Leave')
    return sections, documents


async def load_collections_extra_docs(collection_paths: t.Mapping[str, str]
                                      ) -> t.Mapping[str, CollectionExtraDocsInfoT]:
    '''Load extra docs data.

    :arg collection_paths: Mapping of collection_name to the collection's path.
    :returns: A mapping of collection_name to CollectionExtraDocsInfoT.
    '''
    flog = mlog.fields(func='load_collections_extra_docs')
    flog.debug('Enter')

    loaders = {}
    lib_ctx = app_context.lib_ctx.get()

    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        for collection_name, collection_path in collection_paths.items():
            loaders[collection_name] = await pool.spawn(
                load_collection_extra_docs(collection_name, collection_path))

        responses = await asyncio.gather(*loaders.values())

    # Note: Python dicts have always had a stable order as long as you don't modify the dict.
    # So loaders (implicitly, the keys) and responses have a matching order here.
    result = dict(zip(loaders, responses))

    flog.debug('Leave')
    return result



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Schemas for the plugin DOCUMENTATION data."""

# Ignore Unitialized attribute errors because BaseModel works some magic
# to initialize the attributes when data is loaded into them.
# pyre-ignore-all-errors[13]

import typing as t

import pydantic as p

from .base import (REQUIRED_CLI_F, REQUIRED_ENV_VAR_F, RETURN_TYPE_F,
                   COLLECTION_NAME_F, BaseModel, DeprecationSchema, DocSchema,
                   LocalConfig, OptionsSchema, list_from_scalars, is_json_value,
                   normalize_return_type_names, transform_return_docs)

_SENTINEL = object()


class OptionCliSchema(BaseModel):
    name: str = REQUIRED_CLI_F
    deprecated: DeprecationSchema = p.Field({})
    version_added: str = 'historical'
    version_added_collection: str = COLLECTION_NAME_F


class OptionEnvSchema(BaseModel):
    name: str = REQUIRED_ENV_VAR_F
    deprecated: DeprecationSchema = p.Field({})
    version_added: str = 'historical'
    version_added_collection: str = COLLECTION_NAME_F


class OptionIniSchema(BaseModel):
    key: str
    section: str
    deprecated: DeprecationSchema = p.Field({})
    version_added: str = 'historical'
    version_added_collection: str = COLLECTION_NAME_F


class OptionVarsSchema(BaseModel):
    name: str
    deprecated: DeprecationSchema = p.Field({})
    version_added: str = 'historical'
    version_added_collection: str = COLLECTION_NAME_F


class ReturnSchema(BaseModel):
    """Schema of plugin return data docs."""

    description: t.List[str]
    choices: t.List[str] = []
    elements: str = RETURN_TYPE_F
    returned: str = 'success'
    sample: t.Any = ''  # JSON value
    type: str = RETURN_TYPE_F
    version_added: str = 'historical'
    version_added_collection: str = COLLECTION_NAME_F

    @p.validator('description', pre=True)
    def list_from_scalars(cls, obj):
        return list_from_scalars(obj)

    @p.validator('sample', pre=True)
    def is_json_value(cls, obj):
        if not is_json_value(obj):
            raise ValueError('`sample` must be a JSON value')
        return obj

    @p.validator('type', 'elements', pre=True)
    def normalize_types(cls, obj):
        return normalize_return_type_names(obj)

    @p.root_validator(pre=True)
    def remove_example(cls, values):
        """
        Remove example in favor of sample.

        Having both sample and example is redundant.  Many more plugins are using sample so
        standardize on that.
        """
        example = values.get('example', _SENTINEL)

        if example is not _SENTINEL:
            if values.get('sample'):
                raise ValueError('Cannot specify `example` if `sample` has been specified.')

            if not is_json_value(example):
                raise ValueError('`example` must be a JSON value')

            values['sample'] = example
            del values['example']

        return values


class InnerReturnSchema(ReturnSchema):
    """Nested return schema which allows leaving out description."""

    contains: t.Dict[str, 'InnerReturnSchema'] = {}

    @p.root_validator(pre=True)
    def allow_description_to_be_optional(cls, values):
        # Doing this in a validator so that the json-schema will still flag it as an error
        if 'description' not in values:
            values['description'] = []
        return values


InnerReturnSchema.update_forward_refs()


class OuterReturnSchema(ReturnSchema):
    """Toplevel return schema."""

    contains: t.Dict[str, InnerReturnSchema] = {}


class PluginOptionsSchema(OptionsSchema):
    cli: t.List[OptionCliSchema] = []
    env: t.List[OptionEnvSchema] = []
    ini: t.List[OptionIniSchema] = []
    suboptions: t.Dict[str, 'PluginOptionsSchema'] = {}
    vars: t.List[OptionVarsSchema] = []
    deprecated: DeprecationSchema = p.Field({})


PluginOptionsSchema.update_forward_refs()


class InnerDocSchema(DocSchema):
    options: t.Dict[str, PluginOptionsSchema] = {}


class PluginDocSchema(BaseModel):
    doc: InnerDocSchema


class PluginExamplesSchema(BaseModel):
    examples: str = ''

    @p.validator('examples', pre=True)
    def normalize_examples(cls, value):
        if value is None:
            value = ''
        return value


class PluginMetadataSchema(BaseModel):
    metadata: t.Optional[t.Dict[str, t.Any]] = None


class PluginReturnSchema(BaseModel):
    class Config(LocalConfig):
        fields = {'return_': 'return',
                  }

    return_: t.Dict[str, OuterReturnSchema] = {}

    @p.validator('return_', pre=True)
    def transform_return(cls, obj):
        return transform_return_docs(obj)


class PluginSchema(PluginDocSchema, PluginExamplesSchema, PluginMetadataSchema, PluginReturnSchema,
                   BaseModel):
    """Documentation of an Ansible plugin."""



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Parse documentation from ansible plugins using anible-doc."""

import os
import typing as t


#: Clear Ansible environment variables that set paths where plugins could be found.
ANSIBLE_PATH_ENVIRON: t.Dict[str, str] = os.environ.copy()
ANSIBLE_PATH_ENVIRON.update({'ANSIBLE_COLLECTIONS_PATH': '/dev/null',
                             'ANSIBLE_ACTION_PLUGINS': '/dev/null',
                             'ANSIBLE_CACHE_PLUGINS': '/dev/null',
                             'ANSIBLE_CALLBACK_PLUGINS': '/dev/null',
                             'ANSIBLE_CLICONF_PLUGINS': '/dev/null',
                             'ANSIBLE_CONNECTION_PLUGINS': '/dev/null',
                             'ANSIBLE_FILTER_PLUGINS': '/dev/null',
                             'ANSIBLE_HTTPAPI_PLUGINS': '/dev/null',
                             'ANSIBLE_INVENTORY_PLUGINS': '/dev/null',
                             'ANSIBLE_LOOKUP_PLUGINS': '/dev/null',
                             'ANSIBLE_LIBRARY': '/dev/null',
                             'ANSIBLE_MODULE_UTILS': '/dev/null',
                             'ANSIBLE_NETCONF_PLUGINS': '/dev/null',
                             'ANSIBLE_ROLES_PATH': '/dev/null',
                             'ANSIBLE_STRATEGY_PLUGINS': '/dev/null',
                             'ANSIBLE_TERMINAL_PLUGINS': '/dev/null',
                             'ANSIBLE_TEST_PLUGINS': '/dev/null',
                             'ANSIBLE_VARS_PLUGINS': '/dev/null',
                             'ANSIBLE_DOC_FRAGMENT_PLUGINS': '/dev/null',
                             })
try:
    del ANSIBLE_PATH_ENVIRON['PYTHONPATH']
except KeyError:
    # We just wanted to make sure there was no PYTHONPATH set...
    # all python libs will come from the venv
    pass
try:
    del ANSIBLE_PATH_ENVIRON['ANSIBLE_COLLECTIONS_PATHS']
except KeyError:
    # ANSIBLE_COLLECTIONS_PATHS is the deprecated name replaced by
    # ANSIBLE_COLLECTIONS_PATH
    pass


class ParsingError(Exception):
    """Error raised while parsing plugins for documentation."""


def _get_environment(collection_dir: t.Optional[str]) -> t.Dict[str, str]:
    env = ANSIBLE_PATH_ENVIRON.copy()
    if collection_dir is not None:
        env['ANSIBLE_COLLECTIONS_PATH'] = collection_dir
    else:
        # Copy ANSIBLE_COLLECTIONS_PATH and ANSIBLE_COLLECTIONS_PATHS from the
        # original environment.
        for env_var in ('ANSIBLE_COLLECTIONS_PATH', 'ANSIBLE_COLLECTIONS_PATHS'):
            try:
                del env[env_var]
            except KeyError:
                pass
            if env_var in os.environ:
                env[env_var] = os.environ[env_var]
    return env


class AnsibleCollectionMetadata:
    path: str
    version: t.Optional[str]

    def __init__(self, path: str, version: t.Optional[str]):
        self.path = path
        self.version = version

    def __repr__(self):
        return 'AnsibleCollectionMetadata({0}, {1})'.format(repr(self.path), repr(self.version))

    @classmethod
    def empty(cls, path='.'):
        return cls(path=path, version=None)



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions for working with the ansible-base package."""
import re

import sh


#: Regex to find toplevel directories in tar output
TOPLEVEL_RE: re.Pattern = re.compile('^[^/]+/$')


class InvalidTarball(Exception):
    """Raised when a requested version does not exist."""


async def unpack_tarball(tarname: str, destdir: str) -> str:
    """
    Unpack a tarball.

    :arg tarname: The tarball to unpack.
    :arg destdir: The destination to unpack into.
    :returns: Toplevel of the unpacked directory structure.  This will be
        a subdirectory of `destdir`.
    """
    # FIXME: Need to run tar via run_in_executor()
    # FIXME: Use unpack_tarball for places that are manually calling tar now
    # pyre-ignore[16]
    manifest = sh.tar('-xzvf', tarname, f'-C{destdir}')  # pylint:disable=no-member
    toplevel_dirs = [filename for filename in manifest if TOPLEVEL_RE.match(filename)]

    if len(toplevel_dirs) != 1:
        raise InvalidTarball(f'The tarball {tarname} had more than a single toplevel dir')

    expected_dirname = tarname[:-len('.tar.gz')]
    if toplevel_dirs[0] != expected_dirname:
        raise InvalidTarball(f'The directory in {tarname} was not {expected_dirname}')

    return toplevel_dirs[0]


async def pack_tarball(tarname: str, directory: str) -> str:
    pass



# Copyright: (c) 2019-2020 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)


import json
import os.path

from jinja2 import Environment, FileSystemLoader, PackageLoader

from .filters import do_max, documented_type, html_ify, rst_ify, rst_fmt, rst_xline
from .tests import still_relevant, test_list


# kludge_ns gives us a kludgey way to set variables inside of loops that need to be visible outside
# the loop.  We can get rid of this when we no longer need to build docs with less than Jinja-2.10
# http://jinja.pocoo.org/docs/2.10/templates/#assignments
# With Jinja-2.10 we can use jinja2's namespace feature, restoring the namespace template portion
# of: fa5c0282a4816c4dd48e80b983ffc1e14506a1f5
NS_MAP = {}


def to_kludge_ns(key, value):
    NS_MAP[key] = value
    return ""


def from_kludge_ns(key):
    return NS_MAP[key]


def doc_environment(template_location):
    if isinstance(template_location, str) and os.path.exists(template_location):
        loader = FileSystemLoader(template_location)
    else:
        if isinstance(template_location, str):
            template_pkg = template_location
            template_path = 'templates'
        else:
            template_pkg = template_location[0]
            template_path = template_location[1]

        loader = PackageLoader(template_pkg, template_path)

    env = Environment(loader=loader,
                      variable_start_string="@{",
                      variable_end_string="}@",
                      trim_blocks=True)
    env.globals['xline'] = rst_xline

    # Can be removed (and template switched to use namespace) when we no longer need to build
    # with <Jinja-2.10
    env.globals['to_kludge_ns'] = to_kludge_ns
    env.globals['from_kludge_ns'] = from_kludge_ns
    if 'max' not in env.filters:
        # Jinja < 2.10
        env.filters['max'] = do_max

    if 'tojson' not in env.filters:
        # Jinja < 2.9
        env.filters['tojson'] = json.dumps

    env.filters['rst_ify'] = rst_ify
    env.filters['html_ify'] = html_ify
    env.filters['fmt'] = rst_fmt
    env.filters['xline'] = rst_xline
    env.filters['documented_type'] = documented_type
    env.tests['list'] = test_list
    env.tests['still_relevant'] = still_relevant

    return env



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Entrypoint to the antsibull-docs script."""

from ... import app_context
from ...logging import log
from ...venv import FakeVenvRunner
from .stable import generate_docs_for_all_collections


mlog = log.fields(mod=__name__)


def generate_docs() -> int:
    """
    Create documentation for the current subcommand.

    Current documentation creates documentation for the currently installed version of Ansible,
    as well as the currently installed collections.

    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    flog = mlog.fields(func='generate_docs')
    flog.debug('Begin processing docs')

    app_ctx = app_context.app_ctx.get()

    venv = FakeVenvRunner()

    generate_docs_for_all_collections(
        venv, app_ctx.extra['collection_dir'], app_ctx.extra['dest_dir'],
        breadcrumbs=app_ctx.breadcrumbs)

    return 0



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Output documentation."""

import asyncio
import os.path
import typing as t

import aiofiles
import asyncio_pool

from jinja2 import Template

from . import app_context
from .jinja2.environment import doc_environment
from .logging import log
from .extra_docs import CollectionExtraDocsInfoT
from .docs_parsing import AnsibleCollectionMetadata


mlog = log.fields(mod=__name__)

#: Mapping of plugins to nonfatal errors.  This is the type to use when accepting the plugin.
#: The mapping is of plugin_type: plugin_name: [error_msgs]
PluginErrorsT = t.Mapping[str, t.Mapping[str, t.Sequence[str]]]

#: Mapping to collections to plugins.
#: The mapping is collection_name: plugin_type: plugin_name: plugin_short_description
CollectionInfoT = t.Mapping[str, t.Mapping[str, t.Mapping[str, str]]]

#: Plugins grouped first by plugin type, then by collection
#: The mapping is plugin_type: collection_name: plugin_name: plugin_short_description
PluginCollectionInfoT = t.Mapping[str, t.Mapping[str, t.Mapping[str, str]]]


async def copy_file(source_path: str, dest_path: str) -> None:
    """
    Copy content from one file to another.

    :arg source_path: Source path. Must be a file.
    :arg dest_path: Destination path.
    """
    flog = mlog.fields(func='copy_file')
    flog.debug('Enter')

    lib_ctx = app_context.lib_ctx.get()
    async with aiofiles.open(source_path, 'rb') as f_in:
        async with aiofiles.open(dest_path, 'wb') as f_out:
            # TODO: PY3.8: while chunk := await f.read(lib_ctx.chunksize)
            chunk = await f_in.read(lib_ctx.chunksize)
            while chunk:
                await f_out.write(chunk)
                chunk = await f_in.read(lib_ctx.chunksize)

    flog.debug('Leave')


async def write_plugin_rst(collection_name: str, collection_meta: AnsibleCollectionMetadata,
                           plugin_short_name: str, plugin_type: str,
                           plugin_record: t.Dict[str, t.Any], nonfatal_errors: t.Sequence[str],
                           plugin_tmpl: Template, error_tmpl: Template, dest_dir: str,
                           path_override: t.Optional[str] = None,
                           squash_hierarchy: bool = False) -> None:
    """
    Write the rst page for one plugin.

    :arg collection_name: Dotted colection name.
    :arg collection_meta: Collection metadata object.
    :arg plugin_short_name: short name for the plugin.
    :arg plugin_type: The type of the plugin.  (module, inventory, etc)
    :arg plugin_record: The record for the plugin.  doc, examples, and return are the
        toplevel fields.
    :arg nonfatal_errors: Mapping of plugin to any nonfatal errors that will be displayed in place
        of some or all of the docs
    :arg plugin_tmpl: Template for the plugin.
    :arg error_tmpl: Template to use when there wasn't enough documentation for the plugin.
    :arg dest_dir: Destination directory for the plugin data.  For instance,
        :file:`ansible-checkout/docs/docsite/rst/`.  The directory structure underneath this
        directory will be created if needed.
    :arg squash_hierarchy: If set to ``True``, no directory hierarchy will be used.
                           Undefined behavior if documentation for multiple collections are
                           created.
    """
    flog = mlog.fields(func='write_plugin_rst')
    flog.debug('Enter')

    namespace, collection = collection_name.split('.')
    plugin_name = '.'.join((collection_name, plugin_short_name))

    if not plugin_record:
        flog.fields(plugin_type=plugin_type,
                    plugin_name=plugin_name,
                    nonfatal_errors=nonfatal_errors
                    ).error('{plugin_name} did not return correct DOCUMENTATION.  An error page'
                            ' will be generated.', plugin_name=plugin_name)
        plugin_contents = error_tmpl.render(
            plugin_type=plugin_type, plugin_name=plugin_name,
            collection=collection_name,
            collection_version=collection_meta.version,
            nonfatal_errors=nonfatal_errors)
    else:
        if nonfatal_errors:
            flog.fields(plugin_type=plugin_type,
                        plugin_name=plugin_name,
                        nonfatal_errors=nonfatal_errors
                        ).error('{plugin_name} did not return correct RETURN or EXAMPLES.',
                                plugin_name=plugin_name)
        plugin_contents = plugin_tmpl.render(
            collection=collection_name,
            collection_version=collection_meta.version,
            plugin_type=plugin_type,
            plugin_name=plugin_name,
            doc=plugin_record['doc'],
            examples=plugin_record['examples'],
            returndocs=plugin_record['return'],
            nonfatal_errors=nonfatal_errors)

    if path_override is not None:
        plugin_file = path_override
    else:
        if squash_hierarchy:
            collection_dir = dest_dir
        else:
            collection_dir = os.path.join(dest_dir, 'collections', namespace, collection)
            # This is dangerous but the code that takes dest_dir from the user checks
            # permissions on it to make it as safe as possible.
            os.makedirs(collection_dir, mode=0o755, exist_ok=True)

        plugin_file = os.path.join(collection_dir, f'{plugin_short_name}_{plugin_type}.rst')

    async with aiofiles.open(plugin_file, 'w') as f:
        await f.write(plugin_contents)

    flog.debug('Leave')


async def write_stub_rst(collection_name: str, collection_meta: AnsibleCollectionMetadata,
                         plugin_short_name: str, plugin_type: str,
                         routing_data: t.Mapping[str, t.Any],
                         redirect_tmpl: Template,
                         tombstone_tmpl: Template,
                         dest_dir: str,
                         path_override: t.Optional[str] = None,
                         squash_hierarchy: bool = False) -> None:
    """
    Write the rst page for one plugin stub.

    :arg collection_name: Dotted colection name.
    :arg collection_meta: Collection metadata object.
    :arg plugin_short_name: short name for the plugin.
    :arg plugin_type: The type of the plugin.  (module, inventory, etc)
    :arg routing_data: The routing data record for the plugin stub.  tombstone, deprecation,
        redirect, redirect_is_symlink are the optional toplevel fields.
    :arg redirect_tmpl: Template for redirects.
    :arg tombstone_tmpl: Template for tombstones.
    :arg dest_dir: Destination directory for the plugin data.  For instance,
        :file:`ansible-checkout/docs/docsite/rst/`.  The directory structure underneath this
        directory will be created if needed.
    :arg squash_hierarchy: If set to ``True``, no directory hierarchy will be used.
                           Undefined behavior if documentation for multiple collections are
                           created.
    """
    flog = mlog.fields(func='write_stub_rst')
    flog.debug('Enter')

    namespace, collection = collection_name.split('.')
    plugin_name = '.'.join((collection_name, plugin_short_name))

    if 'tombstone' in routing_data:
        plugin_contents = tombstone_tmpl.render(
            plugin_type=plugin_type,
            plugin_name=plugin_name,
            collection=collection_name,
            collection_version=collection_meta.version,
            tombstone=routing_data['tombstone'])
    else:  # 'redirect' in routing_data
        plugin_contents = redirect_tmpl.render(
            collection=collection_name,
            collection_version=collection_meta.version,
            plugin_type=plugin_type,
            plugin_name=plugin_name,
            redirect=routing_data['redirect'],
            redirect_is_symlink=routing_data.get('redirect_is_symlink') or False,
            deprecation=routing_data.get('deprecation'))

    if path_override is not None:
        plugin_file = path_override
    else:
        if squash_hierarchy:
            collection_dir = dest_dir
        else:
            collection_dir = os.path.join(dest_dir, 'collections', namespace, collection)
            # This is dangerous but the code that takes dest_dir from the user checks
            # permissions on it to make it as safe as possible.
            os.makedirs(collection_dir, mode=0o755, exist_ok=True)

        plugin_file = os.path.join(collection_dir, f'{plugin_short_name}_{plugin_type}.rst')

    async with aiofiles.open(plugin_file, 'w') as f:
        await f.write(plugin_contents)

    flog.debug('Leave')


async def output_all_plugin_rst(collection_to_plugin_info: CollectionInfoT,
                                plugin_info: t.Dict[str, t.Any],
                                nonfatal_errors: PluginErrorsT,
                                dest_dir: str,
                                collection_metadata: t.Mapping[str, AnsibleCollectionMetadata],
                                squash_hierarchy: bool = False) -> None:
    """
    Output rst files for each plugin.

    :arg collection_to_plugin_info: Mapping of collection_name to Mapping of plugin_type to Mapping
        of plugin_name to short_description.
    :arg plugin_info: Documentation information for all of the plugins.
    :arg nonfatal_errors: Mapping of plugins to nonfatal errors.  Using this to note on the docs
        pages when documentation wasn't formatted such that we could use it.
    :arg dest_dir: The directory to place the documentation in.
    :arg collection_metadata: Dictionary mapping collection names to collection metadata objects.
    :arg squash_hierarchy: If set to ``True``, no directory hierarchy will be used.
                           Undefined behavior if documentation for multiple collections are
                           created.
    """
    # Setup the jinja environment
    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    plugin_tmpl = env.get_template('plugin.rst.j2')
    error_tmpl = env.get_template('plugin-error.rst.j2')

    writers = []
    lib_ctx = app_context.lib_ctx.get()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        for collection_name, plugins_by_type in collection_to_plugin_info.items():
            for plugin_type, plugins in plugins_by_type.items():
                for plugin_short_name, dummy_ in plugins.items():
                    plugin_name = '.'.join((collection_name, plugin_short_name))
                    writers.append(await pool.spawn(
                        write_plugin_rst(collection_name,
                                         collection_metadata[collection_name],
                                         plugin_short_name, plugin_type,
                                         plugin_info[plugin_type].get(plugin_name),
                                         nonfatal_errors[plugin_type][plugin_name], plugin_tmpl,
                                         error_tmpl, dest_dir, squash_hierarchy=squash_hierarchy)))

        # Write docs for each plugin
        await asyncio.gather(*writers)


async def output_all_plugin_stub_rst(stubs_info: t.Mapping[
                                         str, t.Mapping[str, t.Mapping[str, t.Any]]],
                                     dest_dir: str,
                                     collection_metadata: t.Mapping[
                                         str, AnsibleCollectionMetadata],
                                     squash_hierarchy: bool = False) -> None:
    """
    Output rst files for each plugin stub.

    :arg stubs_info: Mapping of collection_name to Mapping of plugin_type to Mapping
        of plugin_name to routing information.
    :arg dest_dir: The directory to place the documentation in.
    :arg collection_metadata: Dictionary mapping collection names to collection metadata objects.
    :arg squash_hierarchy: If set to ``True``, no directory hierarchy will be used.
                           Undefined behavior if documentation for multiple collections are
                           created.
    """
    # Setup the jinja environment
    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    redirect_tmpl = env.get_template('plugin-redirect.rst.j2')
    tombstone_tmpl = env.get_template('plugin-tombstone.rst.j2')

    writers = []
    lib_ctx = app_context.lib_ctx.get()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        for collection_name, plugins_by_type in stubs_info.items():
            for plugin_type, plugins in plugins_by_type.items():
                for plugin_short_name, routing_data in plugins.items():
                    writers.append(await pool.spawn(
                        write_stub_rst(collection_name,
                                       collection_metadata[collection_name],
                                       plugin_short_name, plugin_type,
                                       routing_data, redirect_tmpl, tombstone_tmpl,
                                       dest_dir, squash_hierarchy=squash_hierarchy)))

        # Write docs for each plugin
        await asyncio.gather(*writers)


async def write_collection_list(collections: t.Iterable[str], namespaces: t.Iterable[str],
                                template: Template, dest_dir: str,
                                breadcrumbs: bool = True) -> None:
    """
    Write an index page listing all of the collections.

    Each collection will link to an index page listing all content in the collection.

    :arg collections: Iterable of all the collection names.
    :arg namespaces: Iterable of all namespace names.
    :arg template: A template to render the collection index.
    :arg dest_dir: The destination directory to output the index into.
    :kwarg breadcrumbs: Default True.  Set to False if breadcrumbs for collections should be
        disabled.  This will disable breadcrumbs but save on memory usage.
    """
    index_contents = template.render(
        collections=collections, namespaces=namespaces, breadcrumbs=breadcrumbs)
    index_file = os.path.join(dest_dir, 'index.rst')

    async with aiofiles.open(index_file, 'w') as f:
        await f.write(index_contents)


async def write_collection_namespace_index(namespace: str, collections: t.Iterable[str],
                                           template: Template, dest_dir: str,
                                           breadcrumbs: bool = True) -> None:
    """
    Write an index page listing all of the collections for this namespace.

    Each collection will link to an index page listing all content in the collection.

    :arg namespace: The namespace.
    :arg collections: Iterable of all the collection names.
    :arg template: A template to render the collection index.
    :arg dest_dir: The destination directory to output the index into.
    :kwarg breadcrumbs: Default True.  Set to False if breadcrumbs for collections should
        be disabled.  This will disable breadcrumbs but save on memory usage.
    """
    index_contents = template.render(
        namespace=namespace, collections=collections, breadcrumbs=breadcrumbs)
    index_file = os.path.join(dest_dir, 'index.rst')

    async with aiofiles.open(index_file, 'w') as f:
        await f.write(index_contents)


async def write_plugin_type_index(plugin_type: str,
                                  per_collection_plugins: t.Mapping[str, t.Mapping[str, str]],
                                  template: Template,
                                  dest_filename: str) -> None:
    """
    Write an index page for each plugin type.

    :arg plugin_type: The plugin type to write the index for.
    :arg per_collection_plugins: Mapping of collection_name to Mapping of plugin_name to
        short_description.
    :arg template: A template to render the plugin index.
    :arg dest_filename: The destination filename.
    """
    index_contents = template.render(
        plugin_type=plugin_type,
        per_collection_plugins=per_collection_plugins)

    async with aiofiles.open(dest_filename, 'w') as f:
        await f.write(index_contents)


async def write_plugin_lists(collection_name: str,
                             plugin_maps: t.Mapping[str, t.Mapping[str, str]],
                             template: Template,
                             dest_dir: str,
                             collection_meta: AnsibleCollectionMetadata,
                             extra_docs_data: CollectionExtraDocsInfoT,
                             breadcrumbs: bool = True) -> None:
    """
    Write an index page for each collection.

    The per-collection index page links to plugins for each collection.

    :arg plugin_maps: Mapping of plugin_type to Mapping of plugin_name to short_description.
    :arg template: A template to render the collection index.
    :arg dest_dir: The destination directory to output the index into.
    :arg collection_meta: Metadata for the collection.
    :arg extra_docs_data: Extra docs data for the collection.
    :kwarg breadcrumbs: Default True.  Set to False if breadcrumbs for collections should be
        disabled.  This will disable breadcrumbs but save on memory usage.
    """
    index_contents = template.render(
        collection_name=collection_name,
        plugin_maps=plugin_maps,
        collection_version=collection_meta.version,
        breadcrumbs=breadcrumbs,
        extra_docs_sections=extra_docs_data[0])

    # This is only safe because we made sure that the top of the directory tree we're writing to
    # (docs/docsite/rst) is only writable by us.
    os.makedirs(dest_dir, mode=0o755, exist_ok=True)
    index_file = os.path.join(dest_dir, 'index.rst')

    async with aiofiles.open(index_file, 'w') as f:
        await f.write(index_contents)


async def output_collection_index(collection_to_plugin_info: CollectionInfoT,
                                  collection_namespaces: t.Mapping[str, t.List[str]],
                                  dest_dir: str,
                                  breadcrumbs: bool = True) -> None:
    """
    Generate top-level collection index page for the collections.

    :arg collection_to_plugin_info: Mapping of collection_name to Mapping of plugin_type to
        Mapping of plugin_name to short_description.
    :arg collection_namespaces: Mapping from collection namespaces to list of collection names.
    :arg dest_dir: The directory to place the documentation in.
    :kwarg breadcrumbs: Default True.  Set to False if breadcrumbs for collections should be
        disabled.  This will disable breadcrumbs but save on memory usage.
    """
    flog = mlog.fields(func='output_collection_index')
    flog.debug('Enter')

    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    collection_list_tmpl = env.get_template('list_of_collections.rst.j2')

    collection_toplevel = os.path.join(dest_dir, 'collections')
    flog.fields(toplevel=collection_toplevel, exists=os.path.isdir(collection_toplevel)).debug(
        'collection_toplevel exists?')
    # This is only safe because we made sure that the top of the directory tree we're writing to
    # (docs/docsite/rst) is only writable by us.
    os.makedirs(collection_toplevel, mode=0o755, exist_ok=True)

    await write_collection_list(collection_to_plugin_info.keys(), collection_namespaces.keys(),
                                collection_list_tmpl, collection_toplevel, breadcrumbs=breadcrumbs)

    flog.debug('Leave')


async def output_collection_namespace_indexes(collection_namespaces: t.Mapping[str, t.List[str]],
                                              dest_dir: str,
                                              breadcrumbs: bool = True) -> None:
    """
    Generate collection namespace index pages for the collections.

    :arg collection_namespaces: Mapping from collection namespaces to list of collection names.
    :arg dest_dir: The directory to place the documentation in.
    :kwarg breadcrumbs: Default True.  Set to False if breadcrumbs for collections should be
        disabled.  This will disable breadcrumbs but save on memory usage.
    """
    flog = mlog.fields(func='output_collection_namespace_indexes')
    flog.debug('Enter')

    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    collection_list_tmpl = env.get_template('list_of_collections_by_namespace.rst.j2')

    writers = []
    lib_ctx = app_context.lib_ctx.get()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        for namespace, collection_names in collection_namespaces.items():
            namespace_dir = os.path.join(dest_dir, 'collections', namespace)
            # This is only safe because we made sure that the top of the directory tree we're
            # writing to (docs/docsite/rst) is only writable by us.
            os.makedirs(namespace_dir, mode=0o755, exist_ok=True)

            writers.append(await pool.spawn(
                write_collection_namespace_index(
                    namespace, collection_names, collection_list_tmpl, namespace_dir,
                    breadcrumbs=breadcrumbs)))

        await asyncio.gather(*writers)

    flog.debug('Leave')


async def output_plugin_indexes(plugin_info: PluginCollectionInfoT,
                                dest_dir: str) -> None:
    """
    Generate top-level plugin index pages for all plugins of a type in all collections.

    :arg plugin_info: Mapping of plugin_type to Mapping of collection_name to Mapping of
        plugin_name to short_description.
    :arg dest_dir: The directory to place the documentation in.
    """
    flog = mlog.fields(func='output_plugin_indexes')
    flog.debug('Enter')

    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    plugin_list_tmpl = env.get_template('list_of_plugins.rst.j2')

    collection_toplevel = os.path.join(dest_dir, 'collections')
    flog.fields(toplevel=collection_toplevel, exists=os.path.isdir(collection_toplevel)).debug(
        'collection_toplevel exists?')
    # This is only safe because we made sure that the top of the directory tree we're writing to
    # (docs/docsite/rst) is only writable by us.
    os.makedirs(collection_toplevel, mode=0o755, exist_ok=True)

    writers = []
    lib_ctx = app_context.lib_ctx.get()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        for plugin_type, per_collection_data in plugin_info.items():
            filename = os.path.join(
                collection_toplevel, 'index_{type}.rst'.format(type=plugin_type))
            writers.append(await pool.spawn(
                write_plugin_type_index(
                    plugin_type, per_collection_data, plugin_list_tmpl, filename)))

        await asyncio.gather(*writers)

    flog.debug('Leave')


async def output_indexes(collection_to_plugin_info: CollectionInfoT,
                         dest_dir: str,
                         collection_metadata: t.Mapping[str, AnsibleCollectionMetadata],
                         extra_docs_data: t.Mapping[str, CollectionExtraDocsInfoT],
                         squash_hierarchy: bool = False,
                         breadcrumbs: bool = True
                         ) -> None:
    """
    Generate collection-level index pages for the collections.

    :arg collection_to_plugin_info: Mapping of collection_name to Mapping of plugin_type to
        Mapping of plugin_name to short_description.
    :arg dest_dir: The directory to place the documentation in.
    :arg collection_metadata: Dictionary mapping collection names to collection metadata objects.
    :arg extra_docs_data: Dictionary mapping collection names to CollectionExtraDocsInfoT.
    :kwarg squash_hierarchy: If set to ``True``, no directory hierarchy will be used.
        Undefined behavior if documentation for multiple collections are created.
    :kwarg breadcrumbs: Default True.  Set to False if breadcrumbs for collections should be
        disabled.  This will disable breadcrumbs but save on memory usage.
    """
    flog = mlog.fields(func='output_indexes')
    flog.debug('Enter')

    if collection_metadata is None:
        collection_metadata = {}

    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    collection_plugins_tmpl = env.get_template('plugins_by_collection.rst.j2')

    writers = []
    lib_ctx = app_context.lib_ctx.get()

    if not squash_hierarchy:
        collection_toplevel = os.path.join(dest_dir, 'collections')
        flog.fields(toplevel=collection_toplevel, exists=os.path.isdir(collection_toplevel)).debug(
            'collection_toplevel exists?')
        # This is only safe because we made sure that the top of the directory tree we're writing to
        # (docs/docsite/rst) is only writable by us.
        os.makedirs(collection_toplevel, mode=0o755, exist_ok=True)
    else:
        collection_toplevel = dest_dir

    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        for collection_name, plugin_maps in collection_to_plugin_info.items():
            if not squash_hierarchy:
                collection_dir = os.path.join(collection_toplevel, *(collection_name.split('.')))
            else:
                collection_dir = collection_toplevel
            writers.append(await pool.spawn(
                write_plugin_lists(collection_name, plugin_maps, collection_plugins_tmpl,
                                   collection_dir, collection_metadata[collection_name],
                                   extra_docs_data[collection_name],
                                   breadcrumbs=breadcrumbs)))

        await asyncio.gather(*writers)

    flog.debug('Leave')


async def output_extra_docs(dest_dir: str,
                            extra_docs_data: t.Mapping[str, CollectionExtraDocsInfoT],
                            squash_hierarchy: bool = False) -> None:
    """
    Generate collection-level index pages for the collections.

    :arg dest_dir: The directory to place the documentation in.
    :arg extra_docs_data: Dictionary mapping collection names to CollectionExtraDocsInfoT.
    :arg squash_hierarchy: If set to ``True``, no directory hierarchy will be used.
                           Undefined behavior if documentation for multiple collections are
                           created.
    """
    flog = mlog.fields(func='output_extra_docs')
    flog.debug('Enter')

    writers = []
    lib_ctx = app_context.lib_ctx.get()

    if not squash_hierarchy:
        collection_toplevel = os.path.join(dest_dir, 'collections')
    else:
        collection_toplevel = dest_dir

    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        for collection_name, (dummy, documents) in extra_docs_data.items():
            if not squash_hierarchy:
                collection_dir = os.path.join(collection_toplevel, *(collection_name.split('.')))
            else:
                collection_dir = collection_toplevel
            for source_path, rel_path in documents:
                full_path = os.path.join(collection_dir, rel_path)
                os.makedirs(os.path.dirname(full_path), mode=0o755, exist_ok=True)
                writers.append(await pool.spawn(copy_file(source_path, full_path)))

        await asyncio.gather(*writers)

    flog.debug('Leave')



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Schemas for the plugin DOCUMENTATION data."""

import pydantic as p

from .base import BaseModel
from .plugin import InnerDocSchema, PluginExamplesSchema, PluginMetadataSchema, PluginReturnSchema

REQUIRED_CALLBACK_TYPE_F = p.Field(..., regex='^(aggregate|notification|stdout)$')


class InnerCallbackDocSchema(InnerDocSchema):
    """
    Schema describing the structure of callback documentation.

    Differs from other plugins because callbacks have subtypes documented in ``type`` rather than
    having separate types.
    """

    type: str = REQUIRED_CALLBACK_TYPE_F


# Ignore Uninitialized attribute error as BaseModel works some magic to initialize the
# attributes when data is loaded into them.
# pyre-ignore[13]
class CallbackDocSchema(BaseModel):
    doc: InnerCallbackDocSchema


class CallbackSchema(CallbackDocSchema, PluginExamplesSchema, PluginMetadataSchema,
                     PluginReturnSchema, BaseModel):
    """Documentation of callback plugins."""



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Functions for working with the ansible-base package."""
import ast
import asyncio
import os
import re
import tempfile
import typing as t
from functools import lru_cache, partial
from urllib.parse import urljoin

import aiofiles
import sh
from packaging.version import Version as PypiVer

from . import app_context
from .compat import best_get_loop, create_task
from .logging import log
from .utils.http import retry_get

if t.TYPE_CHECKING:
    import aiohttp.client


mlog = log.fields(mod=__name__)

#: URL to checkout ansible-base from.
_ANSIBLE_BASE_URL = str(app_context.AppContext().ansible_base_url)
#: URL to pypi.
_PYPI_SERVER_URL = str(app_context.AppContext().pypi_url)


class UnknownVersion(Exception):
    """Raised when a requested version does not exist."""


class CannotBuild(Exception):
    """Raised when we can't figure out how to build a package."""


class AnsibleBasePyPiClient:
    """Class to retrieve information about AnsibleBase from Pypi."""

    def __init__(self, aio_session: 'aiohttp.client.ClientSession',
                 pypi_server_url: str = _PYPI_SERVER_URL) -> None:
        """
        Initialize the AnsibleBasePypi class.

        :arg aio_session: :obj:`aiohttp.client.ClientSession` to make requests to pypi from.
        :kwarg pypi_server_url: URL to the pypi server to use.
        """
        self.aio_session = aio_session
        self.pypi_server_url = pypi_server_url

    async def _get_json(self, query_url: str) -> t.Dict[str, t.Any]:
        """
        JSON data from a url with retries and return the data as python data structures.
        """
        async with retry_get(self.aio_session, query_url) as response:
            pkg_info = await response.json()
        return pkg_info

    @lru_cache(None)
    async def get_release_info(self) -> t.Dict[str, t.Any]:
        """
        Retrieve information about releases of the ansible-base/ansible-core package from pypi.

        :returns: The dict which represents the release info keyed by version number.
            To examine the data structure, use::

                curl https://pypi.org/pypi/ansible-core/json| python3 -m json.tool

        .. note:: Returns an aggregate of ansible-base and ansible-core releases.
        """
        # Retrieve the ansible-base and ansible-core package info from pypi
        tasks = []
        for package_name in ('ansible-core', 'ansible-base'):
            query_url = urljoin(self.pypi_server_url, f'pypi/{package_name}/json')
            tasks.append(create_task(self._get_json(query_url)))

        # Note: gather maintains the order of results
        results = await asyncio.gather(*tasks)
        release_info = results[1]['releases']  # ansible-base information
        release_info.update(results[0]['releases'])  # ansible-core information

        return release_info

    async def get_versions(self) -> t.List[PypiVer]:
        """
        Get the versions of the ansible-base package on pypi.

        :returns: A list of :pypkg:obj:`packaging.versioning.Version`s
            for all the versions on pypi, including prereleases.
        """
        flog = mlog.fields(func='AnsibleBasePyPiClient.get_versions')
        flog.debug('Enter')

        release_info = await self.get_release_info()
        versions = [PypiVer(r) for r in release_info]
        versions.sort(reverse=True)
        flog.fields(versions=versions).info('sorted list of ansible-core versions')

        flog.debug('Leave')
        return versions

    async def get_latest_version(self) -> PypiVer:
        """
        Get the latest version of ansible-base uploaded to pypi.

        :return: A :pypkg:obj:`packaging.versioning.Version` object representing the latest version
            of the package on pypi.  This may be a pre-release.
        """
        versions = await self.get_versions()
        return versions[0]

    async def retrieve(self, ansible_base_version: str, download_dir: str) -> str:
        """
        Get the release from pypi.

        :arg ansible_base_version: Version of ansible-base to download.
        :arg download_dir: Directory to download the tarball to.
        :returns: The name of the downloaded tarball.
        """
        package_name = get_ansible_core_package_name(ansible_base_version)
        release_info = await self.get_release_info()

        pypi_url = tar_filename = ''
        for release in release_info[ansible_base_version]:
            if release['filename'].startswith(f'{package_name}-{ansible_base_version}.tar.'):
                tar_filename = release['filename']
                pypi_url = release['url']
                break
        else:  # for-else: http://bit.ly/1ElPkyg
            raise UnknownVersion(f'{package_name} {ansible_base_version} does not'
                                 ' exist on {self.pypi_server_url}')

        tar_filename = os.path.join(download_dir, tar_filename)
        async with retry_get(self.aio_session, pypi_url) as response:
            async with aiofiles.open(tar_filename, 'wb') as f:
                lib_ctx = app_context.lib_ctx.get()
                # TODO: PY3.8: while chunk := await response.read(lib_ctx.chunksize):
                chunk = await response.content.read(lib_ctx.chunksize)
                while chunk:
                    await f.write(chunk)
                    chunk = await response.content.read(lib_ctx.chunksize)

        return tar_filename


def get_ansible_core_package_name(ansible_base_version: t.Union[str, PypiVer]) -> str:
    """
    Returns the name of the minimal ansible package.

    :arg ansible_base_version: The version of the minimal ansible package to retrieve the
        name for.
    :returns: 'ansible-core' when the version is 2.11 or higher. Otherwise 'ansible-base'.
    """
    if not isinstance(ansible_base_version, PypiVer):
        ansible_base_version = PypiVer(ansible_base_version)

    if ansible_base_version.major <= 2 and ansible_base_version.minor <= 10:
        return 'ansible-base'

    return 'ansible-core'


def _get_source_version(ansible_base_source: str) -> PypiVer:
    with open(os.path.join(ansible_base_source, 'lib', 'ansible', 'release.py')) as f:
        root = ast.parse(f.read())

    # Find the version of the source
    source_version = None
    # Iterate backwards in case __version__ is assigned to multiple times
    for node in reversed(root.body):
        if isinstance(node, ast.Assign):
            for name in node.targets:
                # These attributes are dynamic so pyre cannot check them
                if name.id == '__version__':  # pyre-ignore[16]
                    source_version = node.value.s  # pyre-ignore[16]
                    break

        if source_version:
            break

    if not source_version:
        raise ValueError('Version was not found')

    return PypiVer(source_version)


def source_is_devel(ansible_base_source: t.Optional[str]) -> bool:
    """
    :arg ansible_base_source: A path to an Ansible-base checkout or expanded sdist or None.
        This will be used instead of downloading an ansible-base package if the version matches
        with ``ansible_base_version``.
    :returns: True if the source looks like it is for the devel branch.
    """
    if ansible_base_source is None:
        return False

    try:
        source_version = _get_source_version(ansible_base_source)
    except Exception:
        return False

    dev_version = re.compile('[.]dev[0-9]+$')
    if dev_version.match(source_version.public):
        return True

    return False


def source_is_correct_version(ansible_base_source: t.Optional[str],
                              ansible_base_version: PypiVer) -> bool:
    """
    :arg ansible_base_source: A path to an Ansible-base checkout or expanded sdist or None.
        This will be used instead of downloading an ansible-base package if the version matches
        with ``ansible_base_version``.
    :arg ansible_base_version: Version of ansible-base to retrieve.
    :returns: True if the source is for a compatible version at or newer than the requested version
    """
    if ansible_base_source is None:
        return False

    try:
        source_version = _get_source_version(ansible_base_source)
    except Exception:
        return False

    # If the source is a compatible version of ansible-base and it is the same or more recent than
    # the requested version then allow this.
    if (source_version.major == ansible_base_version.major
            and source_version.minor == ansible_base_version.minor
            and source_version.micro >= ansible_base_version.micro):
        return True

    return False


@lru_cache(None)
async def checkout_from_git(download_dir: str, repo_url: str = _ANSIBLE_BASE_URL) -> str:
    """
    Checkout the ansible-base git repo.

    :arg download_dir: Directory to checkout into.
    :kwarg: repo_url: The url to the git repo.
    :return: The directory that ansible-base has been checked out to.
    """
    loop = best_get_loop()
    ansible_base_dir = os.path.join(download_dir, 'ansible-base')
    await loop.run_in_executor(None, sh.git, 'clone', repo_url, ansible_base_dir)

    return ansible_base_dir


@lru_cache(None)
async def create_sdist(source_dir: str, dest_dir: str) -> str:
    """
    Create an sdist for the python package at a given path.

    Note that this is not able to create an sdist for any python package.  It has to have a setup.py
    sdist command.

    :arg source_dir: the directory that the python package source is in.
    :arg dest_dir: the directory that the sdist will be written to/
    :returns: path to the sdist.
    """
    loop = best_get_loop()

    # Make sure setup.py exists
    setup_script = os.path.join(source_dir, 'setup.py')
    if not os.path.exists(setup_script):
        raise CannotBuild(f'{source_dir} does not include a setup.py script.  This script cannot'
                          ' build the package')

    # Make a subdir of dest_dir for returning the dist in
    dist_dir_prefix = os.path.join(os.path.basename(source_dir))
    dist_dir = tempfile.mkdtemp(prefix=dist_dir_prefix, dir=dest_dir)

    # execute python setup.py sdist --dist-dir dest_dir/
    # sh maps attributes to commands dynamically so ignore the linting errors there
    # pyre-ignore[16]
    python_cmd = partial(sh.python, _cwd=source_dir)  # pylint:disable=no-member
    try:
        await loop.run_in_executor(None, python_cmd, setup_script, 'sdist', '--dist-dir', dist_dir)
    except Exception as e:
        raise CannotBuild(f'Building {source_dir} failed: {e}')

    dist_files = [f for f in os.listdir(dist_dir) if f.endswith('tar.gz')]
    if len(dist_files) != 1:
        if not dist_files:
            raise CannotBuild(f'Building {source_dir} did not create a tar.gz')

        raise CannotBuild(f'Building {source_dir} created more than one tar.gz files which is not'
                          ' yet supported.')

    return os.path.join(dist_dir, dist_files[0])


async def get_ansible_base(aio_session: 'aiohttp.client.ClientSession',
                           ansible_base_version: str,
                           tmpdir: str,
                           ansible_base_source: t.Optional[str] = None) -> str:
    """
    Create an ansible-base directory of the requested version.

    :arg aio_session: :obj:`aiohttp.client.ClientSession` to make http requests with.
    :arg ansible_base_version: Version of ansible-base to retrieve.  If it is the special string
        ``@devel``, then we will retrieve ansible-base from its git repository.  If it is the
        special string ``@latest``, then we will retrieve the latest version from pypi.
    :arg tmpdir: Temporary directory use as a scratch area for downloading to and the place that the
        ansible-base directory should be placed in.
    :kwarg ansible_base_source: If given, a path to an ansible-base checkout or expanded sdist.
        This will be used instead of downloading an ansible-base package if the version matches
        with ``ansible_base_version``.
    """
    if ansible_base_version == '@devel':
        # is the source usable?
        if source_is_devel(ansible_base_source):
            # source_is_devel() protects against this.  This assert is to inform the type checker
            assert ansible_base_source is not None
            source_location: str = ansible_base_source

        else:
            source_location = await checkout_from_git(tmpdir)

        # Create an sdist from the source that can be installed
        install_file = await create_sdist(source_location, tmpdir)
    else:
        pypi_client = AnsibleBasePyPiClient(aio_session)
        if ansible_base_version == '@latest':
            ansible_base_version: PypiVer = await pypi_client.get_latest_version()
        else:
            ansible_base_version: PypiVer = PypiVer(ansible_base_version)

        # is the source the asked for version?
        if source_is_correct_version(ansible_base_source, ansible_base_version):
            assert ansible_base_source is not None
            # Create an sdist from the source that can be installed
            install_file = await create_sdist(ansible_base_source, tmpdir)
        else:
            install_file = await pypi_client.retrieve(ansible_base_version.public, tmpdir)

    return install_file



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2021

import warnings

# This module is just a backwards compatible location for ansible_doc so the wildcard
# import is just putting all the symbols from docs.ansible_doc into this namespace.
# pylint: disable=wildcard-import,unused-wildcard-import
from .docs.ansible_doc import *  # noqa: F403,F401


warnings.warn('antsibull.schemas.ansible_doc is deprecated.'
              ' Use antsibull.schemas.docs.ansible_doc instead.',
              DeprecationWarning, stacklevel=2)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Entrypoint to the antsibull-docs script."""

import asyncio
import os
import os.path
import tempfile
import typing as t
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor

import aiohttp
import asyncio_pool
from pydantic import ValidationError

from ... import app_context
from ...ansible_base import get_ansible_base
from ...augment_docs import augment_docs
from ...collections import install_together
from ...compat import asyncio_run, best_get_loop
from ...dependency_files import DepsFile
from ...extra_docs import load_collections_extra_docs
from ...docs_parsing.parsing import get_ansible_plugin_info
from ...docs_parsing.fqcn import get_fqcn_parts
from ...docs_parsing.routing import (
    find_stubs,
    load_all_collection_routing,
    remove_redirect_duplicates,
)
from ...galaxy import CollectionDownloader
from ...logging import log
from ...schemas.docs import DOCS_SCHEMAS
from ...venv import VenvRunner, FakeVenvRunner
from ...write_docs import (
    output_all_plugin_rst,
    output_all_plugin_stub_rst,
    output_collection_index,
    output_collection_namespace_indexes,
    output_indexes,
    output_plugin_indexes,
    output_extra_docs,
)
from ...utils.transformations import get_collection_namespaces

if t.TYPE_CHECKING:
    import semantic_version as semver


mlog = log.fields(mod=__name__)

#: Mapping of plugins to nonfatal errors.  This is the type to use when returning the mapping.
PluginErrorsRT = t.DefaultDict[str, t.DefaultDict[str, t.List[str]]]


async def retrieve(ansible_base_version: str,
                   collections: t.Mapping[str, str],
                   tmp_dir: str,
                   galaxy_server: str,
                   ansible_base_source: t.Optional[str] = None,
                   collection_cache: t.Optional[str] = None) -> t.Dict[str, 'semver.Version']:
    """
    Download ansible-core and the collections.

    :arg ansible_base_version: Version of ansible-base/-core to download.
    :arg collections: Map of collection names to collection versions to download.
    :arg tmp_dir: The directory to download into.
    :arg galaxy_server: URL to the galaxy server.
    :kwarg ansible_base_source: If given, a path to an ansible-core checkout or expanded sdist.
        This will be used instead of downloading an ansible-core package if the version matches
        with ``ansible_base_version``.
    :kwarg collection_cache: If given, a path to a directory containing collection tarballs.
        These tarballs will be used instead of downloading new tarballs provided that the
        versions match the criteria (latest compatible version known to galaxy).
    :returns: Map of collection name to directory it is in.  ansible-core will
        use the special key, `_ansible_base`.
    """
    collection_dir = os.path.join(tmp_dir, 'collections')
    os.mkdir(collection_dir, mode=0o700)

    requestors = {}

    lib_ctx = app_context.lib_ctx.get()
    async with aiohttp.ClientSession() as aio_session:
        async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
            requestors['_ansible_base'] = await pool.spawn(
                get_ansible_base(aio_session, ansible_base_version, tmp_dir,
                                 ansible_base_source=ansible_base_source))

            downloader = CollectionDownloader(aio_session, collection_dir,
                                              galaxy_server=galaxy_server,
                                              collection_cache=collection_cache)
            for collection, version in collections.items():
                requestors[collection] = await pool.spawn(
                    downloader.download(collection, version))

            responses = await asyncio.gather(*requestors.values())

    # Note: Python dicts have always had a stable order as long as you don't modify the dict.
    # So requestors (implicitly, the keys) and responses have a matching order here.
    return dict(zip(requestors, responses))


def normalize_plugin_info(plugin_type: str,
                          plugin_info: t.Mapping[str, t.Any]
                          ) -> t.Tuple[t.Dict[str, t.Any], t.List[str]]:
    """
    Normalize and validate all of the plugin docs.

    :arg plugin_type: The type of plugins that we're getting docs for.
    :arg plugin_info: Mapping of plugin_info.  The toplevel keys are plugin names.
        See the schema in :mod:`antsibull.schemas.docs` for what the data should look like and just
        how much conversion we can perform on it.
    :returns: A tuple containing a "copy" of plugin_info with all of the data normalized and a list
        of nonfatal errors.  The plugin_info dict will follow the structure expressed in the schemas
        in :mod:`antsibull.schemas.docs`.  The nonfatal errors are strings representing the problems
        encountered.
    """
    new_info = {}
    errors = []
    # Note: loop through "doc" before any other keys.
    for field in ('doc', 'examples', 'return'):
        try:
            field_model = DOCS_SCHEMAS[plugin_type][field].parse_obj({field: plugin_info[field]})
        except ValidationError as e:
            if field == 'doc':
                # We can't recover if there's not a doc field
                # pydantic exceptions are not picklable (probably due to bugs in the pickle module)
                # so convert it to an exception type which is picklable
                raise ValueError(str(e))

            # But we can use the default value (some variant of "empty") for everything else
            # Note: We looped through doc first and returned an exception if doc did not normalize
            # so we're able to use it in the error message here.
            errors.append(f'Unable to normalize {new_info["doc"]["name"]}: {field}'
                          f' due to: {str(e)}')

            field_model = DOCS_SCHEMAS[plugin_type][field].parse_obj({})

        new_info.update(field_model.dict(by_alias=True))

    return (new_info, errors)


async def normalize_all_plugin_info(plugin_info: t.Mapping[str, t.Mapping[str, t.Any]]
                                    ) -> t.Tuple[t.Dict[str, t.Dict[str, t.Any]], PluginErrorsRT]:
    """
    Normalize the data in plugin_info so that it is ready to be passed to the templates.

    :arg plugin_info: Mapping of information about plugins.  This contains information about all of
        the plugins that are to be documented. See the schema in :mod:`antsibull.schemas.docs` for
        the structure of the information.
    :returns: A tuple of plugin_info (this is a "copy" of the input plugin_info with all of the
        data normalized) and a mapping of errors.  The plugin_info may have less records than the
        input plugin_info if there were plugin records which failed to validate.  The mapping of
        errors takes the form of:

        .. code-block:: yaml

            plugin_type:
                plugin_name:
                    - error string
                    - error string
    """
    loop = best_get_loop()
    lib_ctx = app_context.lib_ctx.get()
    executor = ProcessPoolExecutor(max_workers=lib_ctx.process_max)

    # Normalize each plugin in a subprocess since normalization is CPU bound
    normalizers = {}
    for plugin_type, plugin_list_for_type in plugin_info.items():
        for plugin_name, plugin_record in plugin_list_for_type.items():
            normalizers[(plugin_type, plugin_name)] = loop.run_in_executor(
                executor, normalize_plugin_info, plugin_type, plugin_record)

    results = await asyncio.gather(*normalizers.values(), return_exceptions=True)

    new_plugin_info = defaultdict(dict)
    nonfatal_errors = defaultdict(lambda: defaultdict(list))
    for (plugin_type, plugin_name), plugin_record in zip(normalizers, results):
        # Errors which broke doc parsing (and therefore we won't have enough info to
        # build a docs page)
        if isinstance(plugin_record, Exception):
            # An exception means there is no usable documentation for this plugin
            # Record a nonfatal error and then move on
            nonfatal_errors[plugin_type][plugin_name].append(str(plugin_record))
            continue

        # Errors where we have at least docs.  We can still create a docs page for these with some
        # information left out
        if plugin_record[1]:
            nonfatal_errors[plugin_type][plugin_name].extend(plugin_record[1])

        new_plugin_info[plugin_type][plugin_name] = plugin_record[0]

    return new_plugin_info, nonfatal_errors


def get_plugin_contents(plugin_info: t.Mapping[str, t.Mapping[str, t.Any]],
                        nonfatal_errors: PluginErrorsRT
                        ) -> t.DefaultDict[str, t.DefaultDict[str, t.Dict[str, str]]]:
    """
    Return the collections with their plugins for every plugin type.

    :arg plugin_info: Mapping of plugin type to a mapping of plugin name to plugin record.
        The plugin_type, plugin_name, and short_description from plugin_records are used.
    :arg nonfatal_errors: mapping of plugin type to plugin name to list of error messages.
        The plugin_type and plugin_name are used.
    :returns: A Mapping of plugin type to a mapping of collection name to a mapping of plugin names
        to short_descriptions.
    plugin_type:
        collection:
            - plugin_short_name: short_description
    """
    plugin_contents = defaultdict(lambda: defaultdict(dict))
    # Some plugins won't have an entry in the plugin_info because documentation failed to parse.
    # Those should be documented in the nonfatal_errors information.
    for plugin_type, plugin_list in nonfatal_errors.items():
        for plugin_name, dummy_ in plugin_list.items():
            namespace, collection, short_name = get_fqcn_parts(plugin_name)
            plugin_contents[plugin_type]['.'.join((namespace, collection))][short_name] = ''

    for plugin_type, plugin_list in plugin_info.items():
        for plugin_name, plugin_desc in plugin_list.items():
            namespace, collection, short_name = get_fqcn_parts(plugin_name)
            plugin_contents[plugin_type]['.'.join((namespace, collection))][short_name] = (
                plugin_desc['doc']['short_description'])

    return plugin_contents


def get_collection_contents(plugin_content: t.Mapping[str, t.Mapping[str, t.Mapping[str, str]]],
                            ) -> t.DefaultDict[str, t.Dict[str, t.Mapping[str, str]]]:
    """
    Return the plugins which are in each collection.

    :arg plugin_content: Mapping of plugin type to a mapping of collection name to a mapping of
        plugin name to short description.
    :returns: A Mapping of collection name to a mapping of plugin type to a mapping of plugin names
        to short_descriptions.
    collection:
        plugin_type:
            - plugin_short_name: short_description
    """
    collection_plugins = defaultdict(dict)

    for plugin_type, collection_data in plugin_content.items():
        for collection_name, plugin_data in collection_data.items():
            collection_plugins[collection_name][plugin_type] = plugin_data

    return collection_plugins


def generate_docs_for_all_collections(venv: t.Union[VenvRunner, FakeVenvRunner],
                                      collection_dir: t.Optional[str],
                                      dest_dir: str,
                                      collection_names: t.Optional[t.List[str]] = None,
                                      create_indexes: bool = True,
                                      squash_hierarchy: bool = False,
                                      breadcrumbs: bool = True) -> None:
    """
    Create documentation for a set of installed collections.

    :arg venv: The venv in which ansible-base is installed.
    :arg collection_dir: The directory in which the collections have been installed.
                         If ``None``, the collections are assumed to be in the current
                         search path for Ansible.
    :arg dest_dir: The directory into which the documentation is written.
    :kwarg collection_names: Optional list of collection names. If specified, only documentation
                             for these collections will be collected and generated.
    :kwarg create_indexes: Whether to create the collection, namespace, and plugin indexes. By
                           default, they are created.
    :kwarg squash_hierarchy: If set to ``True``, no directory hierarchy will be used.
                             Undefined behavior if documentation for multiple collections are
                             created.
    :kwarg breadcrumbs: Default True.  Set to False if breadcrumbs for collections should be
        disabled.  This will disable breadcrumbs but save on memory usage.
    """
    flog = mlog.fields(func='generate_docs_for_all_collections')
    flog.notice('Begin')

    # Get the info from the plugins
    plugin_info, collection_metadata = asyncio_run(get_ansible_plugin_info(
        venv, collection_dir, collection_names=collection_names))
    flog.notice('Finished parsing info from plugins and collections')
    # flog.fields(plugin_info=plugin_info).debug('Plugin data')
    # flog.fields(
    #     collection_metadata=collection_metadata).debug('Collection metadata')

    # Load collection routing information
    collection_routing = asyncio_run(load_all_collection_routing(collection_metadata))
    flog.notice('Finished loading collection routing information')
    # flog.fields(collection_routing=collection_routing).debug('Collection routing infos')

    remove_redirect_duplicates(plugin_info, collection_routing)
    stubs_info = find_stubs(plugin_info, collection_routing)
    # flog.fields(stubs_info=stubs_info).debug('Stubs info')

    plugin_info, nonfatal_errors = asyncio_run(normalize_all_plugin_info(plugin_info))
    flog.fields(errors=len(nonfatal_errors)).notice('Finished data validation')
    augment_docs(plugin_info)
    flog.notice('Finished calculating new data')

    # Load collection extra docs data
    extra_docs_data = asyncio_run(load_collections_extra_docs(
        {name: data.path for name, data in collection_metadata.items()}))
    flog.debug('Finished getting collection extra docs data')

    plugin_contents = get_plugin_contents(plugin_info, nonfatal_errors)
    collection_to_plugin_info = get_collection_contents(plugin_contents)
    # Make sure collections without documentable plugins are mentioned
    for collection in collection_metadata:
        collection_to_plugin_info[collection]
    flog.debug('Finished getting collection data')

    collection_namespaces = get_collection_namespaces(collection_to_plugin_info.keys())

    # Only build top-level index if requested
    if create_indexes:
        asyncio_run(output_collection_index(
            collection_to_plugin_info, collection_namespaces, dest_dir, breadcrumbs=breadcrumbs))
        flog.notice('Finished writing collection index')
        asyncio_run(output_collection_namespace_indexes(collection_namespaces, dest_dir,
                                                        breadcrumbs=breadcrumbs))
        flog.notice('Finished writing collection namespace index')
        asyncio_run(output_plugin_indexes(plugin_contents, dest_dir))
        flog.notice('Finished writing plugin indexes')

    asyncio_run(output_indexes(collection_to_plugin_info, dest_dir,
                               collection_metadata=collection_metadata,
                               squash_hierarchy=squash_hierarchy,
                               extra_docs_data=extra_docs_data,
                               breadcrumbs=breadcrumbs))
    flog.notice('Finished writing indexes')

    asyncio_run(output_all_plugin_stub_rst(stubs_info, dest_dir,
                                           collection_metadata=collection_metadata,
                                           squash_hierarchy=squash_hierarchy))
    flog.debug('Finished writing plugin stubs')

    asyncio_run(output_all_plugin_rst(collection_to_plugin_info, plugin_info,
                                      nonfatal_errors, dest_dir,
                                      collection_metadata=collection_metadata,
                                      squash_hierarchy=squash_hierarchy))
    flog.debug('Finished writing plugin docs')

    asyncio_run(output_extra_docs(dest_dir, extra_docs_data,
                                  squash_hierarchy=squash_hierarchy))
    flog.debug('Finished writing extra extra docs docs')


def generate_docs() -> int:
    """
    Create documentation for the stable subcommand.

    Stable documentation creates documentation for a built version of Ansible.  It uses the exact
    versions of collections included in the last Ansible release to generate rst files documenting
    those collections.

    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    flog = mlog.fields(func='generate_docs')
    flog.notice('Begin generating docs')

    app_ctx = app_context.app_ctx.get()

    # Parse the deps file
    flog.fields(deps_file=app_ctx.extra['deps_file']).info('Parse deps file')
    deps_file = DepsFile(app_ctx.extra['deps_file'])
    dummy_, ansible_base_version, collections = deps_file.parse()
    flog.debug('Finished parsing deps file')

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Retrieve ansible-base and the collections
        flog.fields(tmp_dir=tmp_dir).info('created tmpdir')
        collection_tarballs = asyncio_run(
            retrieve(ansible_base_version, collections, tmp_dir,
                     galaxy_server=app_ctx.galaxy_url,
                     ansible_base_source=app_ctx.extra['ansible_base_source'],
                     collection_cache=app_ctx.extra['collection_cache']))
        # flog.fields(tarballs=collection_tarballs).debug('Download complete')
        flog.notice('Finished retrieving tarballs')

        # Get the ansible-core location
        try:
            ansible_base_path = collection_tarballs.pop('_ansible_base')
        except KeyError:
            print('ansible-core did not download successfully')
            return 3
        flog.fields(ansible_base_path=ansible_base_path).info('ansible-core location')

        # Install the collections to a directory

        # Directory that ansible needs to see
        collection_dir = os.path.join(tmp_dir, 'installed')
        # Directory that the collections will be untarred inside of
        collection_install_dir = os.path.join(collection_dir, 'ansible_collections')
        # Safe to recursively mkdir because we created the tmp_dir
        os.makedirs(collection_install_dir, mode=0o700)
        flog.fields(collection_install_dir=collection_install_dir).debug('collection install dir')

        # Install the collections
        asyncio_run(install_together(collection_tarballs.values(), collection_install_dir))
        flog.notice('Finished installing collections')

        # Create venv for ansible-core
        venv = VenvRunner('ansible-core-venv', tmp_dir)
        venv.install_package(ansible_base_path)
        flog.fields(venv=venv).notice('Finished installing ansible-core')

        generate_docs_for_all_collections(venv, collection_dir, app_ctx.extra['dest_dir'],
                                          breadcrumbs=app_ctx.breadcrumbs)

    return 0



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# Author: Felix Fontein <felix@fontein.de>
# License: GPLv3+
# Copyright: Ansible Project, 2020

import json
import os
import traceback
import sys
import typing as t

import sh

from .stable import normalize_plugin_info
from ... import app_context
from ...augment_docs import augment_docs
from ...compat import asyncio_run
from ...docs_parsing import AnsibleCollectionMetadata
from ...docs_parsing.fqcn import get_fqcn_parts, is_fqcn
from ...jinja2.environment import doc_environment
from ...logging import log
from ...vendored.json_utils import _filter_non_json_lines
from ...venv import FakeVenvRunner
from ...write_docs import write_plugin_rst


mlog = log.fields(mod=__name__)


def generate_docs() -> int:
    """
    Create documentation for the current-plugin subcommand.

    Current plugin documentation creates documentation for one currently installed plugin.

    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    flog = mlog.fields(func='generate_docs')
    flog.debug('Begin processing docs')

    app_ctx = app_context.app_ctx.get()
    plugin_type: str = app_ctx.extra['plugin_type']
    plugin_name: str = app_ctx.extra['plugin'][0]

    if not is_fqcn(plugin_name):
        raise NotImplementedError('Priority to implement subcommands is stable, devel, plugin, and'
                                  ' then collection commands. Only the FQCN form is implemented'
                                  ' for the plugin subcommand right now.')

    output_path = os.path.join(app_ctx.extra['dest_dir'], f'{plugin_name}_{plugin_type}.rst')

    try:
        namespace, collection, plugin = get_fqcn_parts(plugin_name)
    except ValueError:
        namespace, collection = 'ansible', 'builtin'
        plugin = plugin_name
    plugin_name = '.'.join([namespace, collection, plugin])

    venv = FakeVenvRunner()
    venv_ansible_doc = venv.get_command('ansible-doc')
    venv_ansible_doc = venv_ansible_doc.bake('-vvv')
    try:
        ansible_doc_results = venv_ansible_doc('-t', plugin_type, '--json', plugin_name)
    except sh.ErrorReturnCode as exc:
        err_msg = []
        formatted_exception = traceback.format_exception(None, exc, exc.__traceback__)
        err_msg.append(f'Exception while parsing documentation for {plugin_type} plugin:'
                       f' {plugin_name}.  Will not document this plugin.')
        err_msg.append(f'Exception:\n{"".join(formatted_exception)}')

        stdout = exc.stdout.decode("utf-8", errors="surrogateescape")
        stderr = exc.stderr.decode("utf-8", errors="surrogateescape")

        err_msg.append(f'Full process stdout:\n{stdout}')
        err_msg.append(f'Full process stderr:\n{stderr}')

        sys.stderr.write('\n'.join(err_msg))
        return 1

    stdout = ansible_doc_results.stdout.decode("utf-8", errors="surrogateescape")

    plugin_info = json.loads(_filter_non_json_lines(stdout)[0])[plugin_name]
    flog.debug('Finished parsing info from plugin')

    plugin_info, errors = normalize_plugin_info(plugin_type, plugin_info)
    flog.debug('Finished normalizing data')

    # The cast is needed to make pyre happy. It seems to not being able to
    # understand that
    #     t.Dict[str, t.Dict[str, t.Dict[str, typing.Any]]]
    # is acceptable for
    #     t.MutableMapping[str, t.MutableMapping[str, typing.Any]].
    augment_docs(t.cast(t.MutableMapping[str, t.MutableMapping[str, t.Any]], {
        plugin_type: {
            plugin_name: plugin_info
        }
    }))

    # Setup the jinja environment
    env = doc_environment(('antsibull.data', 'docsite'))
    # Get the templates
    plugin_tmpl = env.get_template('plugin.rst.j2')
    error_tmpl = env.get_template('plugin-error.rst.j2')

    asyncio_run(write_plugin_rst(
        '.'.join([namespace, collection]), AnsibleCollectionMetadata.empty(), plugin, plugin_type,
        plugin_info, errors, plugin_tmpl, error_tmpl, '',
        path_override=output_path))
    flog.debug('Finished writing plugin docs')

    return 0



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Parse documentation from ansible plugins using anible-doc."""

import asyncio
import json
import sys
import os
import traceback
import typing as t
from concurrent.futures import ThreadPoolExecutor

import sh

from .. import app_context
from ..compat import best_get_loop, create_task
from ..constants import DOCUMENTABLE_PLUGINS
from ..logging import log
from ..vendored.json_utils import _filter_non_json_lines
from .fqcn import get_fqcn_parts
from . import _get_environment, ParsingError, AnsibleCollectionMetadata

if t.TYPE_CHECKING:
    from ..venv import VenvRunner, FakeVenvRunner


mlog = log.fields(mod=__name__)


def _process_plugin_results(plugin_type: str,
                            plugin_names: t.Iterable[str],
                            plugin_info: t.Sequence[t.Union[sh.RunningCommand, Exception]]
                            ) -> t.Dict:
    """
    Process the results from running ansible-doc.

    In particular, log errors and remove them from the output.

    :arg plugin_type: The type of plugin.  See :attr:`DOCUMENTABLE_PLUGINS` for a list
        of allowed types.
    :arg plugin_names: Iterable of the plugin_names that were processed in the same order as the
        plugin_info.
    :arg plugin_info: List of results running sh.ansible_doc on each plugin.
    :returns: Dictionary mapping plugin_name to the results from ansible-doc.
    """
    flog = mlog.fields(func='_process_plugin_results')
    flog.debug('Enter')

    results = {}
    for plugin_name, ansible_doc_results in zip(plugin_names, plugin_info):
        plugin_log = flog.fields(plugin_type=plugin_type, plugin_name=plugin_name)

        if isinstance(ansible_doc_results, Exception):
            error_fields = {}
            error_fields['exception'] = traceback.format_exception(
                None, ansible_doc_results, ansible_doc_results.__traceback__)

            if isinstance(ansible_doc_results, sh.ErrorReturnCode):
                error_fields['stdout'] = ansible_doc_results.stdout.decode(
                    'utf-8', errors='surrogateescape')
                error_fields['stderr'] = ansible_doc_results.stderr.decode(
                    'utf-8', errors='surrogateescape')

            plugin_log.fields(**error_fields).error(
                'Exception while parsing documentation.  Will not document this plugin.')
            continue

        stdout = ansible_doc_results.stdout.decode('utf-8', errors='surrogateescape')

        # ansible-doc returns plugins shipped with ansible-base using no namespace and collection.
        # For now, we fix these entries to use the ansible.builtin collection here.  The reason we
        # do it here instead of as part of a general normalization step is that other plugins
        # (site-specific ones from ANSIBLE_LIBRARY, for instance) will also be returned with no
        # collection name.  We know that we don't have any of those in this code (because we set
        # ANSIBLE_LIBRARY and other plugin path variables to /dev/null) so we can safely fix this
        # here but not outside the ansible-doc backend.
        fqcn = plugin_name
        try:
            get_fqcn_parts(fqcn)
        except ValueError:
            fqcn = f'ansible.builtin.{plugin_name}'

        try:
            ansible_doc_output = json.loads(_filter_non_json_lines(stdout)[0])
        except Exception as e:
            formatted_exception = traceback.format_exception(None, e, e.__traceback__)

            plugin_log.fields(ansible_doc_stdout=stdout, exception=formatted_exception,
                              traceback=traceback.format_exc()).error(
                                  'ansible-doc did not return json data.'
                                  ' Will not document this plugin.')
            continue

        results[fqcn] = ansible_doc_output[plugin_name]

    return results


async def _get_plugin_info(plugin_type: str, ansible_doc: 'sh.Command',
                           max_workers: int,
                           collection_names: t.Optional[t.List[str]] = None) -> t.Dict[str, t.Any]:
    """
    Retrieve info about all Ansible plugins of a particular type.

    :arg plugin_type: The type of plugin.  See :attr:`DOCUMENTABLE_PLUGINS` for a list
        of allowed types.
    :arg ansible_doc: An :sh:obj:`sh.Command` object that will run the ansible-doc command.
        This command should already have been baked with any necessary environment and
        common arguments.
        :returns: A nested dictionary structure that looks like::
            Dictionary mapping to plugin information.  The dictionary looks like::

                plugin_name:  # This is the canonical name for the plugin and includes
                              # namespace.collection_name unless the plugin comes from
                              # ansible-base.
                    {information from ansible-doc --json.  See the ansible-doc documentation for
                     more info.}
    :kwarg max_workers: The maximum number of threads that should be run in parallel by this
        function.
    :returns: Mapping of fqcn's to plugin_info.
    """
    flog = mlog.fields(func='_get_plugin_info')
    flog.debug('Enter')

    # Get the list of plugins
    ansible_doc_list_cmd_list = ['--list', '--t', plugin_type, '--json']
    if collection_names and len(collection_names) == 1:
        # Ansible-doc list allows to filter by one collection
        ansible_doc_list_cmd_list.append(collection_names[0])
    ansible_doc_list_cmd = ansible_doc(*ansible_doc_list_cmd_list)
    raw_plugin_list = ansible_doc_list_cmd.stdout.decode('utf-8', errors='surrogateescape')
    # Note: Keep ansible_doc_list_cmd around until we know if we need to use it in an error message.
    plugin_map = json.loads(_filter_non_json_lines(raw_plugin_list)[0])
    del raw_plugin_list
    del ansible_doc_list_cmd

    # Filter plugin map
    if collection_names is not None:
        prefixes = ['{name}.'.format(name=collection) for collection in collection_names]
        plugin_map = {
            key: value for key, value in plugin_map.items()
            if any(key.startswith(prefix) for prefix in prefixes)
        }

    loop = best_get_loop()

    # For each plugin, get its documentation
    extractors = {}
    executor = ThreadPoolExecutor(max_workers=max_workers)
    for plugin_name in plugin_map.keys():
        extractors[plugin_name] = loop.run_in_executor(executor, ansible_doc, '-t', plugin_type,
                                                       '--json', plugin_name)
    plugin_info = await asyncio.gather(*extractors.values(), return_exceptions=True)

    results = _process_plugin_results(plugin_type, extractors, plugin_info)

    flog.debug('Leave')
    return results


def get_collection_metadata(venv: t.Union['VenvRunner', 'FakeVenvRunner'],
                            env: t.Dict[str, str],
                            collection_names: t.Optional[t.List[str]] = None,
                            ) -> t.Dict[str, AnsibleCollectionMetadata]:
    collection_metadata = {}

    # Obtain ansible.builtin version
    if collection_names is None or 'ansible.builtin' in collection_names:
        venv_ansible = venv.get_command('ansible')
        ansible_version_cmd = venv_ansible('--version', _env=env)
        raw_result = ansible_version_cmd.stdout.decode('utf-8', errors='surrogateescape')
        path: t.Optional[str] = None
        version: t.Optional[str] = None
        for line in raw_result.splitlines():
            if line.strip().startswith('ansible python module location'):
                path = line.split('=', 2)[1].strip()
            if line.startswith('ansible '):
                version = line[len('ansible '):]
        collection_metadata['ansible.builtin'] = AnsibleCollectionMetadata(
            path=path, version=version)

    # Obtain collection versions
    venv_ansible_galaxy = venv.get_command('ansible-galaxy')
    ansible_collection_list_cmd = venv_ansible_galaxy('collection', 'list', _env=env)
    raw_result = ansible_collection_list_cmd.stdout.decode('utf-8', errors='surrogateescape')
    current_base_path = None
    for line in raw_result.splitlines():
        parts = line.split()
        if len(parts) >= 2:
            if parts[0] == '#':
                current_base_path = parts[1]
            else:
                collection_name = parts[0]
                version = parts[1]
                if '.' in collection_name:
                    if collection_names is None or collection_name in collection_names:
                        namespace, name = collection_name.split('.', 2)
                        collection_metadata[collection_name] = AnsibleCollectionMetadata(
                            path=os.path.join(current_base_path, namespace, name),
                            version=None if version == '*' else version)

    return collection_metadata


async def get_ansible_plugin_info(venv: t.Union['VenvRunner', 'FakeVenvRunner'],
                                  collection_dir: t.Optional[str],
                                  collection_names: t.Optional[t.List[str]] = None
                                  ) -> t.Tuple[
                                      t.Mapping[str, t.Mapping[str, t.Any]],
                                      t.Mapping[str, AnsibleCollectionMetadata]]:
    """
    Retrieve information about all of the Ansible Plugins.

    :arg venv: A VenvRunner into which Ansible has been installed.
    :arg collection_dir: Directory in which the collections have been installed.
                         If ``None``, the collections are assumed to be in the current
                         search path for Ansible.
    :arg collection_names: Optional list of collections. If specified, will only collect
                           information for plugins in these collections.
    :returns: An tuple. The first component is a nested directory structure that looks like:

            plugin_type:
                plugin_name:  # Includes namespace and collection.
                    {information from ansible-doc --json.  See the ansible-doc documentation
                     for more info.}

        The second component is a Mapping of collection names to metadata.
    """
    flog = mlog.fields(func='get_ansible_plugin_info')
    flog.debug('Enter')

    env = _get_environment(collection_dir)

    # Setup an sh.Command to run ansible-doc from the venv with only the collections we
    # found as providers of extra plugins.

    venv_ansible_doc = venv.get_command('ansible-doc')
    venv_ansible_doc = venv_ansible_doc.bake('-vvv', _env=env)

    # We invoke _get_plugin_info once for each documentable plugin type.  Within _get_plugin_info,
    # new threads are spawned to handle waiting for ansible-doc to parse files and give us results.
    # To keep ourselves under thread_max, we need to divide the number of threads we're allowed over
    # each call of _get_plugin_info.
    # Why use thread_max instead of process_max?  Even though this ultimately invokes separate
    # ansible-doc processes, the limiting factor is IO as ansible-doc reads from disk.  So it makes
    # sense to scale up to thread_max instead of process_max.

    # Allocate more for modules because the vast majority of plugins are modules
    lib_ctx = app_context.lib_ctx.get()
    module_workers = max(int(.7 * lib_ctx.thread_max), 1)
    other_workers = int((lib_ctx.thread_max - module_workers) / (len(DOCUMENTABLE_PLUGINS) - 1))
    if other_workers < 1:
        other_workers = 1

    extractors = {}
    for plugin_type in DOCUMENTABLE_PLUGINS:
        if plugin_type == 'module':
            max_workers = module_workers
        else:
            max_workers = other_workers
        extractors[plugin_type] = create_task(
            _get_plugin_info(plugin_type, venv_ansible_doc, max_workers, collection_names))

    results = await asyncio.gather(*extractors.values(), return_exceptions=True)

    plugin_map = {}
    err_msg = []
    an_exception = None
    for plugin_type, extraction_result in zip(extractors, results):

        if isinstance(extraction_result, Exception):
            an_exception = extraction_result
            formatted_exception = traceback.format_exception(None, extraction_result,
                                                             extraction_result.__traceback__)
            err_msg.append(f'Exception while parsing documentation for {plugin_type} plugins')
            err_msg.append(f'Exception:\n{"".join(formatted_exception)}')

        # Note: Exception will also be True.
        if isinstance(extraction_result, sh.ErrorReturnCode):
            stdout = extraction_result.stdout.decode("utf-8", errors="surrogateescape")
            stderr = extraction_result.stderr.decode("utf-8", errors="surrogateescape")
            err_msg.append(f'Full process stdout:\n{stdout}')
            err_msg.append(f'Full process stderr:\n{stderr}')

        if err_msg:
            sys.stderr.write('\n'.join(err_msg))
            sys.stderr.write('\n')
            continue

        plugin_map[plugin_type] = extraction_result

    if an_exception:
        # We wanted to print out all of the exceptions raised by parsing the output but once we've
        # done so, we want to then fail by raising one of the exceptions.
        raise ParsingError('Parsing of plugins failed')

    flog.debug('Retrieving collection metadata')
    collection_metadata = get_collection_metadata(venv, env, collection_names)

    flog.debug('Leave')
    return (plugin_map, collection_metadata)



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Constant values for use throughout the antsibull codebase."""

from typing import FrozenSet


#: All the types of ansible plugins
PLUGIN_TYPES: FrozenSet[str] = frozenset(('become', 'cache', 'callback', 'cliconf', 'connection',
                                          'httpapi', 'inventory', 'lookup', 'shell', 'strategy',
                                          'vars', 'module', 'module_utils',))

#: The subset of PLUGINS which we build documentation for
DOCUMENTABLE_PLUGINS: FrozenSet[str] = frozenset(('become', 'cache', 'callback', 'cliconf',
                                                  'connection', 'httpapi', 'inventory', 'lookup',
                                                  'netconf', 'shell', 'vars', 'module',
                                                  'strategy',))



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""
Functions for parsing fully qualified collection names.

Some definitions:

:Fully qualified collection names: aka FQCN are the complete name for a collection.  They
    include the namespace, collection name, and the plugin name.  If a plugin is within a python
    package within the collection dir, then the FQCN may include that as well.
:short_name: The pieces of the fqcn that don't include the namespace and collection name.  If the
    `collections:` keyword is used in an ansible playbook to get shorter names, the short name is
    what users would have to specify.
"""

import re
import typing as t


#: Format that a collection namespace and collection name must follow
NAMESPACE_RE_STR = '[a-z0-9][a-z0-9_]+'
#: Format of a FQCN
FQCN_RE = re.compile(fr'({NAMESPACE_RE_STR})\.({NAMESPACE_RE_STR})\.(.*)')
FQCN_STRICT_RE = re.compile(
    fr'({NAMESPACE_RE_STR})\.({NAMESPACE_RE_STR})\.({NAMESPACE_RE_STR}(?:\.{NAMESPACE_RE_STR})*)')

# FQCN_RE and FQCN_STRICT_RE match certain Fully Qualified Collection Names. FQCN_RE is more liberal
# than FQCN_STRICT_RE and allows random characters after the namespace and collection name, while
# FQCN_STRICT_RE is closer to the definition in Ansible. We use FQCN the same as the term FQCR
# (Fully Qualified Collection Reference) is used in Ansible.
#
# The set of possible FQCRs accepted by Ansible is defined in
# https://github.com/ansible/ansible/blob/devel/lib/ansible/utils/collection_loader/_collection_finder.py#L662
# similarly to what we define as FQCN_STRICT_RE, except that it allows arbitrary words matching
# '[a-zA-Z0-9_]+' instead of words matching NAMESPACE_RE_STR.
#
# The following table shows some differences and similarities:
#
# String               FQCN_RE  FQCN_STRICT_RE  Ansible FQCR
# ----------------------------------------------------------
# 'a.b.c'              yes      yes             yes
# 'a.b.c.d'            yes      yes             yes
# 'a.b.c-d'            yes      no              no
# 'A.B.C'              no       no              yes
# 'a..c'               no       no              no


def get_fqcn_parts(fqcn: str) -> t.Tuple[str, str, str]:
    """
    Parse a fqcn into its three parts.

    :arg fqcn: The fqcn string.
    :returns: A tuple of (namespace, collection name, plugin short_name).
    :raises ValueError: If the string could not be parsed into fqcn components.
    """
    match = FQCN_RE.match(fqcn)
    if not match:
        raise ValueError(f'{fqcn} could not be parsed into fqcn parts')
    return match.groups()


def is_fqcn(value: str) -> bool:
    """
    Return whether ``value`` is a Fully Qualified Collection Name (FQCN).

    :arg value: The value to test.
    :returns: ``True`` if the value is a FQCN, ``False`` if it is not.
    """
    return bool(FQCN_STRICT_RE.match(value))



from packaging.version import  Version
import pytest

import antsibull.ansible_base as ab


@pytest.mark. parametrize('version', ('2.9', '2.9.10', '1.0', '1', '0.7', '2.10', '2.10.0',
                                      '2.10.8', '2.10.12'))
def test_get_core_package_name_returns_ansible_base(version):
    assert ab.get_ansible_core_package_name(version) == 'ansible-base'
    assert ab.get_ansible_core_package_name(Version(version)) == 'ansible-base'


@pytest.mark.parametrize('version', ('2.11', '2.11.0', '2.11.8', '2.11.12', '2.13',
                                     '2.11.0a1', '3', '3.7.10'))
def test_get_core_package_name_returns_ansible_core(version):
    assert ab.get_ansible_core_package_name(version) == 'ansible-core'
    assert ab.get_ansible_core_package_name(Version(version)) == 'ansible-core'



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Parse documentation from ansible plugins using anible-doc."""

import typing as t

from .. import app_context
from ..logging import log
from .ansible_doc import get_ansible_plugin_info as ansible_doc_get_ansible_plugin_info
from .ansible_internal import get_ansible_plugin_info as ansible_internal_get_ansible_plugin_info
from . import AnsibleCollectionMetadata

if t.TYPE_CHECKING:
    from ..venv import VenvRunner, FakeVenvRunner


mlog = log.fields(mod=__name__)


async def get_ansible_plugin_info(venv: t.Union['VenvRunner', 'FakeVenvRunner'],
                                  collection_dir: t.Optional[str],
                                  collection_names: t.Optional[t.List[str]] = None
                                  ) -> t.Tuple[
                                    t.Mapping[str, t.Mapping[str, t.Any]],
                                    t.Mapping[str, AnsibleCollectionMetadata]]:
    """
    Retrieve information about all of the Ansible Plugins.

    :arg venv: A VenvRunner into which Ansible has been installed.
    :arg collection_dir: Directory in which the collections have been installed.
                         If ``None``, the collections are assumed to be in the current
                         search path for Ansible.
    :arg collection_names: Optional list of collections. If specified, will only collect
                           information for plugins in these collections.
    :returns: An tuple. The first component is a nested directory structure that looks like:

            plugin_type:
                plugin_name:  # Includes namespace and collection.
                    {information from ansible-doc --json.  See the ansible-doc documentation
                     for more info.}

        The second component is a Mapping of collection names to metadata.

    """
    lib_ctx = app_context.lib_ctx.get()

    doc_parsing_backend = lib_ctx.doc_parsing_backend
    if doc_parsing_backend == 'ansible-internal':
        return await ansible_internal_get_ansible_plugin_info(
            venv, collection_dir, collection_names=collection_names)
    if doc_parsing_backend == 'ansible-doc':
        return await ansible_doc_get_ansible_plugin_info(
            venv, collection_dir, collection_names=collection_names)

    raise Exception('Invalid value for doc_parsing_backend: %s' % (doc_parsing_backend, ))



import pytest

from antsibull import dependency_files as df

PIECES = """
community.general
# This is a comment
community.aws
    # Supported by ansible
    ansible.posix
    ansible.windows
# Third parties
purestorage.flasharray
"""

PARSED_PIECES = ['community.general', 'community.aws', 'ansible.posix', 'ansible.windows', 'purestorage.flasharray']

DEPS = """
_ansible_version: 2.10.5
# this is a comment
_ansible_base_version: 2.10.4
    # supported by ansible
    ansible.netcommon: 1.4.1
    ansible.posix: 1.1.1
    ansible.windows: 1.3.0
# third parties
dellemc.os10: 1.0.2
"""

PARSED_DEPS_ANSIBLE_VERSION = '2.10.5'
PARSED_DEPS_ANSIBLE_BASE_VERSION = '2.10.4'
PARSED_DEPS_DEPS = {
    'ansible.netcommon': '1.4.1',
    'ansible.posix': '1.1.1',
    'ansible.windows': '1.3.0',
    'dellemc.os10': '1.0.2'
}

BUILD = """
_ansible_version: 2.10
# this is a comment
_ansible_base_version: 2.10.1
    # supported by ansible
    ansible.netcommon: >=1.2.0,<2.0.0
    ansible.posix: >=1.1.0,<2.0.0
    ansible.windows: >=1.0.0,<2.0.0
# third parties
dellemc.os10: >=1.0.0,<1.1.0
"""

PARSED_BUILD_ANSIBLE_VERSION = '2.10'
PARSED_BUILD_ANSIBLE_BASE_VERSION = '2.10.1'
PARSED_BUILD_DEPS = {
    'ansible.netcommon': '>=1.2.0,<2.0.0',
    'ansible.posix': '>=1.1.0,<2.0.0',
    'ansible.windows': '>=1.0.0,<2.0.0',
    'dellemc.os10': '>=1.0.0,<1.1.0'
}

def test_parse_pieces(tmp_path):
    pieces_filename = tmp_path / 'pieces.in'
    with open(pieces_filename, 'w') as f:
        f.write(PIECES)
    assert df.parse_pieces_file(pieces_filename) == PARSED_PIECES

def test_deps_file_parse(tmp_path):
    deps_filename = tmp_path / 'deps.in'
    with open(deps_filename, 'w') as f:
        f.write(DEPS)
    parsed_deps = df.DepsFile(deps_filename).parse()
    assert parsed_deps.ansible_version == PARSED_DEPS_ANSIBLE_VERSION
    assert parsed_deps.ansible_base_version == PARSED_DEPS_ANSIBLE_BASE_VERSION
    assert parsed_deps.deps == PARSED_DEPS_DEPS

def test_build_file_parse(tmp_path):
    build_filename = tmp_path / 'build.in'
    with open(build_filename, 'w') as f:
        f.write(BUILD)
    parsed_build = df.DepsFile(build_filename).parse()
    assert parsed_build.ansible_version == PARSED_BUILD_ANSIBLE_VERSION
    assert parsed_build.ansible_base_version == PARSED_BUILD_ANSIBLE_BASE_VERSION
    assert parsed_build.deps == PARSED_BUILD_DEPS



# Author: Felix Fontein <felix@fontein.de>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""General functions for working with aiohttp."""

import asyncio
import typing as t
import warnings

import aiohttp

from .. import app_context
from ..logging import log


mlog = log.fields(mod=__name__)


def _format_call(command: str, args: t.Tuple[t.Any, ...], kwargs: t.Mapping[str, t.Any]) -> str:
    arguments = [repr(a) for a in args] + ['{0}={1}'.format(k, repr(v)) for k, v in kwargs.items()]
    return 'aio_session.{command}({args})'.format(
        command=command,
        args=', '.join(arguments),
    )


class RetryGetManager:
    response: t.Optional[aiohttp.ClientResponse]

    def __init__(self,
                 aio_session: 'aiohttp.client.ClientSession',
                 args: t.Tuple[t.Any, ...],
                 kwargs: t.Mapping[str, t.Any],
                 max_retries: int,
                 acceptable_error_codes: t.Iterable[int],
                 ):
        self.aio_session = aio_session
        self.args = args
        self.kwargs = kwargs
        self.max_retries = max_retries
        self.acceptable_error_codes = acceptable_error_codes
        self.call_string = _format_call('get', args, kwargs)
        self.response = None

    async def __aenter__(self) -> aiohttp.ClientResponse:
        flog = mlog.fields(func='RetryGetManager.__aenter__')
        flog.debug('Enter')

        error_codes = []
        for retry in range(self.max_retries):
            flog.debug('Execute {0}', self.call_string)
            try:
                response = await self.aio_session.get(*self.args, **self.kwargs)
                status_code = response.status
                flog.debug('Status code {0}'.format(status_code))
                if status_code < 400 or status_code in self.acceptable_error_codes:
                    self.response = response
                    flog.debug('Leave')
                    return response
                error_codes.append(status_code)
                response.close()
            except Exception as error:
                flog.trace()
                status_code = str(error)
                error_codes.append(status_code)

            failed = retry + 1 == self.max_retries
            warnings.warn('{0} failed with status code {1}{2}'.format(
                self.call_string,
                status_code,
                ', finally failed.' if failed else ', retrying...'
            ))
            if failed:
                break

            await asyncio.sleep(retry * 0.5)

        flog.debug('Raise error')
        raise Exception('Repeated error when calling {0}: received status codes {1}'.format(
            self.call_string, ', '.join([str(error) for error in error_codes])))

    async def __aexit__(self, exc_type, exc, tb) -> None:
        flog = mlog.fields(func='RetryGetManager.__aexit__')
        flog.debug('Enter')
        response = self.response
        self.response = None
        if response is not None:
            response.close()
        flog.debug('Leave')


def retry_get(aio_session: 'aiohttp.client.ClientSession',
              *args,
              acceptable_error_codes: t.Optional[t.Iterable[int]] = None,
              max_retries: t.Optional[int] = None,
              **kwargs) -> t.AsyncContextManager[aiohttp.ClientResponse]:
    # Handle default value for max_retries
    lib_ctx = app_context.lib_ctx.get()
    if max_retries is None:
        max_retries = lib_ctx.max_retries

    # Make sure max_retries is at least 1
    max_retries = max(max_retries, 1)

    # pyre-ignore[7]: no idea how to make pyre see that RetryGetManager has correct type
    return RetryGetManager(aio_session, args, kwargs, max_retries, acceptable_error_codes or ())



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Entrypoint to the antsibull-docs script."""

import argparse
import os
import os.path
import stat
import sys
from typing import Callable, Dict, List

import twiggy

from ..logging import log, initialize_app_logging
initialize_app_logging()

# We have to call initialize_app_logging() before these imports so that the log object is configured
# correctly before other antisbull modules make copies of it.
# pylint: disable=wrong-import-position
from .. import app_context  # noqa: E402
from ..args import (  # noqa: E402
    InvalidArgumentError, get_toplevel_parser, normalize_toplevel_options
)
from ..compat import BooleanOptionalAction  # noqa: E402
from ..config import load_config  # noqa: E402
from ..constants import DOCUMENTABLE_PLUGINS  # noqa: E402
from ..filesystem import UnableToCheck, writable_via_acls  # noqa: E402
from ..docs_parsing.fqcn import is_fqcn  # noqa: E402
from .doc_commands import collection, current, devel, plugin, stable  # noqa: E402
# pylint: enable=wrong-import-position


mlog = log.fields(mod=__name__)

#: Mapping from command line subcommand names to functions which implement those
#: The functions need to take a single argument, the processed list of args.
ARGS_MAP: Dict[str, Callable] = {'devel': devel.generate_docs,
                                 'stable': stable.generate_docs,
                                 'current': current.generate_docs,
                                 'collection': collection.generate_docs,
                                 'plugin': plugin.generate_docs,
                                 }

#: The filename for the file which lists raw collection names
DEFAULT_PIECES_FILE: str = 'ansible.in'


def _normalize_docs_options(args: argparse.Namespace) -> None:
    args.dest_dir = os.path.abspath(os.path.realpath(args.dest_dir))

    # We're going to be writing a deep hierarchy of files into this directory so we need to make
    # sure that the user understands that this needs to be a directory which has been secured
    # against malicious usage:

    # Exists already
    try:
        stat_results = os.stat(args.dest_dir)

        if not stat.S_ISDIR(stat_results.st_mode):
            raise FileNotFoundError()
    except FileNotFoundError:
        raise InvalidArgumentError(f'{args.dest_dir} must be an existing directory owned by you,'
                                   f' and only be writable by the owner')

    # Owned by the user
    euid = os.geteuid()
    if stat_results[stat.ST_UID] != euid:
        raise InvalidArgumentError(f'{args.dest_dir} must be owned by you, and only be writable'
                                   f' by the owner')

    # Writable only by the user
    if stat.S_IMODE(stat_results.st_mode) & (stat.S_IWOTH | stat.S_IWGRP):
        raise InvalidArgumentError(f'{args.dest_dir} must only be writable by the owner')

    try:
        if writable_via_acls(args.dest_dir, euid):
            raise InvalidArgumentError(f'Filesystem acls grant write on {args.dest_dir} to'
                                       ' additional users')
    except UnableToCheck:
        # We've done our best but some systems don't even have acls on their filesystem so we can't
        # error here.
        pass


def _normalize_devel_options(args: argparse.Namespace) -> None:
    if args.command != 'devel':
        return

    if not os.path.isfile(args.pieces_file):
        raise InvalidArgumentError(f'The pieces file, {args.pieces_file}, must already exist.'
                                   ' It should contain one namespace.collection per line')


def _normalize_stable_options(args: argparse.Namespace) -> None:
    if args.command != 'stable':
        return

    if not os.path.isfile(args.deps_file):
        raise InvalidArgumentError(f'The deps file, {args.deps_file}, must already exist.'
                                   ' It should contain one namespace.collection with version'
                                   ' per line')


def _normalize_collection_options(args: argparse.Namespace) -> None:
    if args.command != 'collection':
        return

    if args.squash_hierarchy and len(args.collections) > 1:
        raise InvalidArgumentError('The option --squash-hierarchy can only be used when'
                                   ' only one collection is specified')


def _normalize_current_options(args: argparse.Namespace) -> None:
    if args.command != 'current':
        return

    if args.collection_dir is not None:
        if not os.path.isdir(os.path.join(args.collection_dir, 'ansible_collections')):
            raise InvalidArgumentError(f'The collection directory, {args.collection_dir}, must be'
                                       ' a directory containing a subdirectory ansible_collections')


def _normalize_plugin_options(args: argparse.Namespace) -> None:
    if args.command != 'plugin':
        return

    for plugin_name in args.plugin:
        if not is_fqcn(plugin_name) and not os.path.isfile(plugin_name):
            raise InvalidArgumentError(f'The plugin, {plugin_name}, must be an existing file,'
                                       f' or it must be a FQCN.')


def parse_args(program_name: str, args: List[str]) -> argparse.Namespace:
    """
    Parse and coerce the command line arguments.

    :arg program_name: The name of the program
    :arg args: A list of the command line arguments
    :returns: A :python:obj:`argparse.Namespace`
    :raises InvalidArgumentError: Whenever there's something wrong with the arguments.
    """
    flog = mlog.fields(func='parse_args')
    flog.fields(program_name=program_name, raw_args=args).info('Enter')

    # Overview of parsers:
    # * docs_parser is an abstract parser.  Contains options that all of the antisbull-docs
    #   subcommands use.
    # * cache_parser is a mixin for subcommands which operate on the ansible-core sources and
    #   therefore they can use a preinstalled version of the code instead of downloading it
    #   themselves.
    # * whole_site_parser is a mixin for subcommands which can choose to build a structure
    #   for integration into a comprehensive website.
    # * devel, stable, current, collection, file: These parsers contain all of the options for those
    #   respective subcommands.
    docs_parser = argparse.ArgumentParser(add_help=False)
    docs_parser.add_argument('--dest-dir', default='.',
                             help='Directory to write the output to')

    cache_parser = argparse.ArgumentParser(add_help=False)
    # TODO: Remove --ansible-base-cache once the ansible/ansible docs-build test is updated
    cache_parser.add_argument('--ansible-base-source', '--ansible-base-cache', default=None,
                              help='Checkout or expanded tarball of the ansible-base package.  If'
                              ' this is a git checkout it must be the HEAD of the branch you are'
                              ' building for. If it is an expanded tarball, the __version__ will'
                              ' be checked to make sure it is compatible with and the same or'
                              ' later version than requested by the deps file.')
    cache_parser.add_argument('--collection-cache', default=None,
                              help='Directory of collection tarballs.  These will be used instead'
                              ' of downloading fresh versions provided that they meet the criteria'
                              ' (Latest version of the collections known to galaxy).')

    whole_site_parser = argparse.ArgumentParser(add_help=False)
    whole_site_parser.add_argument('--breadcrumbs', '--gretel',
                                   dest='breadcrumbs', action=BooleanOptionalAction,
                                   default=argparse.SUPPRESS,
                                   help='Determines whether to add breadcrumbs to plugin docs via'
                                   ' hidden sphinx toctrees. This can take up a significant'
                                   ' amount of memory if there are a large number of plugins so'
                                   ' you can disable this if necessary. (default: True)')
    whole_site_parser.add_argument('--indexes',
                                   dest='indexes', action=BooleanOptionalAction,
                                   default=argparse.SUPPRESS,
                                   help='Determines whether to create the collection index and'
                                   ' plugin indexes. They may not be needed if you have'
                                   ' a different structure for your website. (default: True)')
    # --skip-indexes is for backwards compat.  We want all negations to be --no-* in the future.
    whole_site_parser.add_argument('--skip-indexes',
                                   dest='indexes', action='store_false',
                                   default=argparse.SUPPRESS,
                                   help='Do not create the collection index and plugin indexes.'
                                   ' This option is deprecated in favor of --no-indexes')

    parser = get_toplevel_parser(prog=program_name,
                                 description='Script to manage generated documentation for'
                                 ' ansible')
    subparsers = parser.add_subparsers(title='Subcommands', dest='command',
                                       help='for help use: `SUBCOMMANDS -h`')
    subparsers.required = True

    #
    # Document the next version of ansible
    #
    devel_parser = subparsers.add_parser('devel',
                                         parents=[docs_parser, cache_parser, whole_site_parser],
                                         description='Generate documentation for the next major'
                                         ' release of Ansible')
    devel_parser.add_argument('--pieces-file', default=DEFAULT_PIECES_FILE,
                              help='File containing a list of collections to include')

    #
    # Document a released version of ansible
    #
    stable_parser = subparsers.add_parser('stable',
                                          parents=[docs_parser, cache_parser, whole_site_parser],
                                          description='Generate documentation for a current'
                                          ' version of ansible')
    stable_parser.add_argument('--deps-file', required=True,
                               help='File which contains the list of collections and'
                               ' versions which were included in this version of Ansible')

    #
    # Document the currently installed version of ansible
    #
    current_parser = subparsers.add_parser('current',
                                           parents=[docs_parser],
                                           description='Generate documentation for the current'
                                           ' installed version of ansible and the current installed'
                                           ' collections')
    current_parser.add_argument('--collection-dir',
                                help='Path to the directory containing ansible_collections. If not'
                                ' specified, all collections in the currently configured ansible'
                                ' search paths will be used')

    #
    # Document one or more specified collections
    #
    collection_parser = subparsers.add_parser('collection',
                                              parents=[docs_parser, whole_site_parser],
                                              description='Generate documentation for specified'
                                              ' collections')
    collection_parser.add_argument('--collection-version', default='@latest',
                                   help='The version of the collection to document.  The special'
                                   ' version, "@latest" can be used to download and document the'
                                   ' latest version from galaxy.')
    collection_parser.add_argument('--use-current', action='store_true',
                                   help='Assumes that all arguments are collection names, and'
                                   ' these collections have been installed with the current'
                                   ' version of ansible. Specified --collection-version will be'
                                   ' ignored.')
    collection_parser.add_argument('--squash-hierarchy', action='store_true',
                                   help='Do not use the full hierarchy collections/namespace/name/'
                                   ' in the destination directory. Only valid if there is only'
                                   ' one collection specified.  Implies --no-indexes.')
    collection_parser.add_argument(nargs='+', dest='collections',
                                   help='One or more collections to document.  If the names are'
                                   ' directories on disk, they will be parsed as expanded'
                                   ' collections. Otherwise, if they could be collection'
                                   ' names, they will be downloaded from galaxy.')

    #
    # Document a specifically named plugin
    #
    file_parser = subparsers.add_parser('plugin',
                                        parents=[docs_parser],
                                        description='Generate documentation for a single plugin')
    file_parser.add_argument(nargs=1, dest='plugin', action='store',
                             help='A single file to document. Either a path to a file, or a FQCN.'
                             ' In the latter case, the plugin is assumed to be installed for'
                             ' the current ansible version.')
    file_parser.add_argument('--plugin-type', action='store', default='module',
                             choices=DOCUMENTABLE_PLUGINS,
                             help='The type of the plugin')

    flog.debug('Argument parser setup')

    if '--ansible-base-cache' in args:
        flog.warning('The CLI parameter, `--ansible-base-cache` has been renamed to'
                     ' `--ansible-base-source.  Please use that instead')

    if '--skip-indexes' in args:
        flog.warning('The CLI parameter, `--skip-indexes` has been renamed to'
                     ' `--no-indexes`.  Please use that instead')
        if '--indexes' or '--no-indexes' in args:
            raise InvalidArgumentError('You cannot use `--indexes`/`--no-indexes` with'
                                       ' `--skip-indexes`. Please remove `--skip-indexes`'
                                       ' and try again.')

    args: argparse.Namespace = parser.parse_args(args)
    flog.fields(args=args).debug('Arguments parsed')

    # Validation and coercion
    normalize_toplevel_options(args)
    _normalize_docs_options(args)
    _normalize_devel_options(args)
    _normalize_stable_options(args)
    _normalize_current_options(args)
    _normalize_collection_options(args)
    _normalize_plugin_options(args)
    flog.fields(args=args).debug('Arguments normalized')

    # Note: collections aren't validated as existing files or collection names here because talking
    # to galaxy to validate the collection names goes beyond the scope of what parsing and
    # validating the command line should do.

    return args


def run(args: List[str]) -> int:
    """
    Run the program.

    :arg args: A list of command line arguments.  Typically :python:`sys.argv`.
    :returns: A program return code.  0 for success, integers for any errors.  These are documented
        in :func:`main`.
    """
    flog = mlog.fields(func='run')
    flog.fields(raw_args=args).info('Enter')

    program_name = os.path.basename(args[0])
    try:
        args: argparse.Namespace = parse_args(program_name, args[1:])
    except InvalidArgumentError as e:
        flog.error(e)
        print(e)
        return 2
    flog.fields(args=args).info('Arguments parsed')

    cfg = load_config(args.config_file)
    flog.fields(config=cfg).info('Config loaded')

    context_data = app_context.create_contexts(args=args, cfg=cfg)
    with app_context.app_and_lib_context(context_data) as (app_ctx, dummy_):
        twiggy.dict_config(app_ctx.logging_cfg.dict())
        flog.debug('Set logging config')

        flog.fields(command=args.command).info('Action')
        return ARGS_MAP[args.command]()


def main() -> int:
    """
    Entrypoint called from the script.

    console_scripts call functions which take no parameters.  However, it's hard to test a function
    which takes no parameters so this function lightly wraps :func:`run`, which actually does the
    heavy lifting.

    :returns: A program return code.

    Return codes:
        :0: Success
        :1: Unhandled error.  See the Traceback for more information.
        :2: There was a problem with the command line arguments
        :3: Unexpected problem downloading ansible-base
    """
    return run(sys.argv)



# coding: utf-8
# Author: Felix Fontein <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2021

import os
import os.path
import re
import typing as t

import docutils.utils
import rstcheck

from .extra_docs import (
    find_extra_docs,
    lint_required_conditions,
    load_extra_docs_index,
    ExtraDocsIndexError,
)
from .yaml import load_yaml_file


_RST_LABEL_DEFINITION = re.compile(r'''^\.\. _([^:]+):''')


def load_collection_name(path_to_collection: str) -> str:
    '''Load collection name (namespace.name) from collection's galaxy.yml.'''
    galaxy_yml_path = os.path.join(path_to_collection, 'galaxy.yml')
    if not os.path.isfile(galaxy_yml_path):
        raise Exception(f'Cannot find file {galaxy_yml_path}')

    galaxy_yml = load_yaml_file(galaxy_yml_path)
    collection_name = '{namespace}.{name}'.format(**galaxy_yml)
    return collection_name


def lint_optional_conditions(content: str, path: str, collection_name: str
                             ) -> t.List[t.Tuple[int, int, str]]:
    '''Check a extra docs RST file's content for whether it satisfied the required conditions.

    Return a list of errors.
    '''
    results = rstcheck.check(
        content, filename=path,
        report_level=docutils.utils.Reporter.WARNING_LEVEL)
    return [(result[0], 0, result[1]) for result in results]


def lint_collection_extra_docs_files(path_to_collection: str
                                     ) -> t.List[t.Tuple[str, int, int, str]]:
    try:
        collection_name = load_collection_name(path_to_collection)
    except Exception:
        return [(
            path_to_collection, 0, 0, 'Cannot identify collection with galaxy.yml at this path')]
    result = []
    all_labels = set()
    docs = find_extra_docs(path_to_collection)
    for doc in docs:
        try:
            # Load content
            with open(doc[0], 'r', encoding='utf-8') as f:
                content = f.read()
            # Rstcheck
            errors = lint_optional_conditions(content, doc[0], collection_name)
            result.extend((doc[0], line, col, msg) for (line, col, msg) in errors)
            # Lint labels
            labels, errors = lint_required_conditions(content, collection_name)
            all_labels.update(labels)
            result.extend((doc[0], line, col, msg) for (line, col, msg) in errors)
        except Exception as e:
            result.append((doc[0], 0, 0, str(e)))
    index_path = os.path.join(path_to_collection, 'docs', 'docsite', 'extra-docs.yml')
    try:
        sections, errors = load_extra_docs_index(index_path)
        result.extend((index_path, 0, 0, error) for error in errors)
    except ExtraDocsIndexError as exc:
        if len(docs) > 0:
            # Only report the missing index_path as an error if we found documents
            result.append((index_path, 0, 0, str(exc)))
    return result



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020

import asyncio
import os
import os.path
import tempfile
import typing as t

import aiohttp
import asyncio_pool

from .stable import generate_docs_for_all_collections
from ... import app_context
from ...ansible_base import get_ansible_base
from ...collections import install_together
from ...compat import asyncio_run
from ...dependency_files import parse_pieces_file
from ...galaxy import CollectionDownloader, DownloadResults
from ...logging import log
from ...venv import VenvRunner

if t.TYPE_CHECKING:
    import semantic_version as semver


mlog = log.fields(mod=__name__)


async def retrieve(collections: t.List[str],
                   tmp_dir: str,
                   galaxy_server: str,
                   ansible_base_source: t.Optional[str] = None,
                   collection_cache: t.Optional[str] = None) -> t.Dict[str, 'semver.Version']:
    """
    Download ansible-core and the latest versions of the collections.

    :arg collections: List of collection names to download.
    :arg tmp_dir: The directory to download into.
    :arg galaxy_server: URL to the galaxy server.
    :kwarg ansible_base_source: If given, a path to an ansible-core checkout or expanded sdist.
        This will be used instead of downloading an ansible-core package if the version matches
        with ``ansible_base_version``.
    :kwarg collection_cache: If given, a path to a directory containing collection tarballs.
        These tarballs will be used instead of downloading new tarballs provided that the
        versions match the criteria (latest compatible version known to galaxy).
    :returns: Map of collection name to directory it is in.  ansible-core will
        use the special key, `_ansible_base`.
    """
    collection_dir = os.path.join(tmp_dir, 'collections')
    os.mkdir(collection_dir, mode=0o700)

    requestors = {}

    lib_ctx = app_context.lib_ctx.get()
    async with aiohttp.ClientSession() as aio_session:
        async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
            requestors['_ansible_base'] = await pool.spawn(
                get_ansible_base(aio_session, '@devel', tmp_dir,
                                 ansible_base_source=ansible_base_source))

            downloader = CollectionDownloader(aio_session, collection_dir,
                                              galaxy_server=galaxy_server,
                                              collection_cache=collection_cache)
            for collection in collections:
                requestors[collection] = await pool.spawn(
                    downloader.download_latest_matching(collection, '*'))

            responses = await asyncio.gather(*requestors.values())

    responses = [
        data.download_path if isinstance(data, DownloadResults) else data for data in responses
    ]
    # Note: Python dicts have always had a stable order as long as you don't modify the dict.
    # So requestors (implicitly, the keys) and responses have a matching order here.
    return dict(zip(requestors, responses))


def generate_docs() -> int:
    """
    Create documentation for the devel subcommand.

    Devel documentation creates documentation for the current development version of Ansible.
    It uses the latest collection releases for the collections mentioned in the specified pieces
    file to generate rst files documenting those collections.

    :returns: A return code for the program.  See :func:`antsibull.cli.antsibull_docs.main` for
        details on what each code means.
    """
    flog = mlog.fields(func='generate_docs')
    flog.notice('Begin generating docs')

    app_ctx = app_context.app_ctx.get()

    # Parse the pieces file
    flog.fields(deps_file=app_ctx.extra['pieces_file']).info('Parse pieces file')
    collections = parse_pieces_file(app_ctx.extra['pieces_file'])
    flog.debug('Finished parsing deps file')

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Retrieve ansible-base and the collections
        flog.fields(tmp_dir=tmp_dir).info('created tmpdir')
        collection_tarballs = asyncio_run(
            retrieve(collections, tmp_dir,
                     galaxy_server=app_ctx.galaxy_url,
                     ansible_base_source=app_ctx.extra['ansible_base_source'],
                     collection_cache=app_ctx.extra['collection_cache']))
        # flog.fields(tarballs=collection_tarballs).debug('Download complete')
        flog.notice('Finished retrieving tarballs')

        # Get the ansible-core location
        try:
            ansible_base_path = collection_tarballs.pop('_ansible_base')
        except KeyError:
            print('ansible-core did not download successfully')
            return 3
        flog.fields(ansible_base_path=ansible_base_path).info('ansible-core location')

        # Install the collections to a directory

        # Directory that ansible needs to see
        collection_dir = os.path.join(tmp_dir, 'installed')
        # Directory that the collections will be untarred inside of
        collection_install_dir = os.path.join(collection_dir, 'ansible_collections')
        # Safe to recursively mkdir because we created the tmp_dir
        os.makedirs(collection_install_dir, mode=0o700)
        flog.fields(collection_install_dir=collection_install_dir).debug('collection install dir')

        # Install the collections
        asyncio_run(install_together(collection_tarballs.values(), collection_install_dir))
        flog.notice('Finished installing collections')

        # Create venv for ansible-core
        venv = VenvRunner('ansible-core-venv', tmp_dir)
        venv.install_package(ansible_base_path)
        flog.fields(venv=venv).notice('Finished installing ansible-core')

        generate_docs_for_all_collections(venv, collection_dir, app_ctx.extra['dest_dir'],
                                          breadcrumbs=app_ctx.breadcrumbs)

    return 0



# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Helper to use pkgutil.get_data without having to check the return value."""

import pkgutil


def get_antsibull_data(filename: str) -> bytes:
    '''
    Retrieve data from the antsibull.data package as bytes.

    The filename can be a relative path separated with '/' to access subdirectories.
    See https://docs.python.org/3/library/pkgutil.html#pkgutil.get_data for details.
    '''
    data = pkgutil.get_data('antsibull.data', filename)
    if data is None:
        raise RuntimeError(f"Cannot find {filename} in the antsibull.data package")
    return data



# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Parse documentation from ansible plugins using anible-doc."""

import json
import tempfile
import typing as t

from ..logging import log
from ..utils.get_pkg_data import get_antsibull_data
from ..vendored.json_utils import _filter_non_json_lines
from . import _get_environment, AnsibleCollectionMetadata

if t.TYPE_CHECKING:
    from ..venv import VenvRunner, FakeVenvRunner


mlog = log.fields(mod=__name__)


async def get_ansible_plugin_info(venv: t.Union['VenvRunner', 'FakeVenvRunner'],
                                  collection_dir: t.Optional[str],
                                  collection_names: t.Optional[t.List[str]] = None
                                  ) -> t.Tuple[
                                      t.Mapping[str, t.Mapping[str, t.Any]],
                                      t.Mapping[str, AnsibleCollectionMetadata]]:
    """
    Retrieve information about all of the Ansible Plugins.

    :arg venv: A VenvRunner into which Ansible has been installed.
    :arg collection_dir: Directory in which the collections have been installed.
                         If ``None``, the collections are assumed to be in the current
                         search path for Ansible.
    :arg collection_names: Optional list of collections. If specified, will only collect
                           information for plugins in these collections.
    :returns: An tuple. The first component is a nested directory structure that looks like:

            plugin_type:
                plugin_name:  # Includes namespace and collection.
                    {information from ansible-doc --json.  See the ansible-doc documentation
                     for more info.}

        The second component is a Mapping of collection names to metadata.
    """
    flog = mlog.fields(func='get_ansible_plugin_info')
    flog.debug('Enter')

    env = _get_environment(collection_dir)

    venv_python = venv.get_command('python')

    with tempfile.NamedTemporaryFile() as tmp_file:
        tmp_file.write(get_antsibull_data('collection-enum.py'))
        collection_enum_args = [tmp_file.name]
        if collection_names is not None:
            # Script allows to filter by collections
            collection_enum_args.extend(collection_names)
        collection_enum_cmd = venv_python(*collection_enum_args, _env=env)
        raw_result = collection_enum_cmd.stdout.decode('utf-8', errors='surrogateescape')
        result = json.loads(_filter_non_json_lines(raw_result)[0])
        del raw_result
        del collection_enum_cmd

    plugin_map = {}
    for plugin_type, plugins in result['plugins'].items():
        plugin_map[plugin_type] = {}
        for plugin_name, plugin_data in plugins.items():
            if '.' not in plugin_name:
                plugin_name = 'ansible.builtin.{0}'.format(plugin_name)
            if 'ansible-doc' in plugin_data:
                plugin_map[plugin_type][plugin_name] = plugin_data['ansible-doc']
            else:
                plugin_log = flog.fields(plugin_type=plugin_type, plugin_name=plugin_name)
                plugin_log.fields(error=plugin_data['error']).error(
                    'Error while extracting documentation. Will not document this plugin.')

    collection_metadata = {}
    for collection_name, collection_data in result['collections'].items():
        collection_metadata[collection_name] = AnsibleCollectionMetadata(
            path=collection_data['path'],
            version=collection_data.get('version'))

    flog.debug('Leave')
    return (plugin_map, collection_metadata)



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Argument parsing helpers."""

import argparse
import os.path

from .compat import metadata


class InvalidArgumentError(Exception):
    """A problem parsing or validating a command line argument."""


def get_toplevel_parser(**kwargs) -> argparse.ArgumentParser:
    """
    Create a toplevel argument parser with options common across all scripts.

    :args kwargs: This function takes any keyword arguments and passes them directly on to
        the :class:`argparse.ArgumentParser` constructor.
    :returns: :class:`argparse.ArgumentParser` with common script arguments added.
    """
    try:
        version = metadata.version('antsibull')
    except metadata.PackageNotFoundError:
        # If there's no metadata foun, assume we're running from source
        version = 'source'

    toplevel_parser = argparse.ArgumentParser(**kwargs)
    toplevel_parser.add_argument('--version', action='version', version=version,
                                 help='Print the antsibull version')
    toplevel_parser.add_argument('--config-file', default=[], action='append',
                                 help='Specify one or more config files to use to configure the'
                                 ' program. If more than one are specified, keys from later'
                                 ' config files override keys from earlier ones.')
    return toplevel_parser


def normalize_toplevel_options(args: argparse.Namespace) -> None:
    """
    Normalize and validate the common cli arguments.

    :arg args: The argparse parsed arguments.  The arguments added by the common parser will be
        validated and normalized.

    .. warning:: This function operates by side effect.

        Any normalization needed will be applied directly to ``args``.
    """
    for conf_file in args.config_file:
        if not os.path.isfile(conf_file):
            raise InvalidArgumentError(f'The user specified config file, {conf_file},'
                                       ' must exist.')



# coding: utf-8
# Author: Felix Fontein <felix@fontein.de>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2021
"""General transformation functions."""

import typing as t
from collections import defaultdict


def get_collection_namespaces(collection_names: t.Iterable[str]) -> t.Dict[str, t.List[str]]:
    """
    Return the plugins which are in each collection.

    :arg collection_names: An iterable of collection names.
    :returns: Mapping from collection namespaces to list of collection names.
    """
    namespaces = defaultdict(list)
    for collection_name in collection_names:
        namespace, name = collection_name.split('.', 1)
        namespaces[namespace].append(name)
    return namespaces



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020

import json
import os.path
import tempfile

import sh
from jinja2 import Template

from . import app_context
from .dependency_files import DepsFile
from .utils.get_pkg_data import get_antsibull_data


def build_collection_command():
    app_ctx = app_context.app_ctx.get()
    with tempfile.TemporaryDirectory() as working_dir:
        collection_dir = os.path.join(working_dir, 'community', 'ansible')

        sh.ansible_galaxy('collection', 'init', 'community.ansible', '--init-path', working_dir)
        # Copy the README.md file
        readme = get_antsibull_data('README_md.txt')
        with open(os.path.join(collection_dir, 'README.md'), 'wb') as f:
            f.write(readme)

        # Parse the deps file
        deps_file = DepsFile(os.path.join(app_ctx.extra['data_dir'], app_ctx.extra['deps_file']))
        dummy1_, dummy2_, deps = deps_file.parse()

        # Template the galaxy.yml file
        dep_string = json.dumps(deps)
        dep_string.replace(', ', ',\n    ')
        galaxy_yml = get_antsibull_data('galaxy_yml.j2').decode('utf-8')
        galaxy_yml_tmpl = Template(galaxy_yml)
        galaxy_yml_contents = galaxy_yml_tmpl.render(version=app_ctx.extra['ansible_version'],
                                                     dependencies=dep_string)

        with open(os.path.join(collection_dir, 'galaxy.yml'), 'w') as f:
            f.write(galaxy_yml_contents)

        sh.ansible_galaxy('collection', 'build',
                          '--output-path', app_ctx.extra['collection_dir'],
                          collection_dir)

    return 0
